import copy
import itertools
import sys
from collections import defaultdict

import numpy as np
import pytest
from pytest import fixture, mark

from negmas.apps.scml import ManufacturingProfile, Product, Process, InputOutput, Job, RunningCommandInfo
from negmas.apps.scml.simulators import SlowFactorySimulator, FastFactorySimulator, storage_as_array
from negmas.apps.scml.world import Factory
from negmas.apps.scml.common import NO_PRODUCTION

n_lines = 5
n_levels = 4
n_products = 5
n_processes = 3
initial_wallet = 100.0
n_steps = 100
max_storage = sys.maxsize
initial_storage = dict(zip(range(n_products * n_levels), range(10, 10 * (n_products * n_levels), 10)))


@fixture(scope='module')
def products():
    return [Product(id=i + l * n_levels, production_level=l, name=f'{l}_{i}', catalog_price=(i + 1) * (l + 1),
                    expires_in=None)
            for i in range(n_products) for l in range(n_levels)]


@fixture(scope='module')
def processes():
    return [Process(id=i + l * n_levels, production_level=l, name=f'p{l}_{i}'
                    , inputs={InputOutput(product=i + (l - 1) * n_levels, quantity=3, step=0.0)
            , InputOutput(product=i + (l - 1) * n_levels, quantity=2, step=0.0)}
                    , outputs={InputOutput(product=i + l * n_levels, quantity=1, step=1.0)
            , InputOutput(product=i + l * n_levels, quantity=2, step=1.0)}
                    , historical_cost=1 + i + l * n_levels)
            for i in range(n_processes) for l in range(1, n_levels)]


@fixture(scope='module')
def profiles(processes):
    return [ManufacturingProfile(n_steps=i + 1, cost=10 * (i + 1), initial_pause_cost=1, running_pause_cost=2
                                 , resumption_cost=3, cancellation_cost=4, line=l, process=p)
            for i, (p, l) in enumerate(zip(processes, itertools.cycle(range(n_lines))))]


@fixture(scope='module')
def n_profiles(profiles):
    return len(profiles)


@fixture
def empty_factory(profiles):
    return Factory(id='factory', profiles=profiles, initial_wallet=initial_wallet, initial_storage={}
                   , max_storage=max_storage)


@fixture
def factory_with_storage(profiles):
    return Factory(id='factory', profiles=profiles, initial_storage=copy.deepcopy(initial_storage)
                   , initial_wallet=initial_wallet, max_storage=max_storage)


@fixture
def slow_simulator(profiles, products):
    return SlowFactorySimulator(initial_wallet=initial_wallet, initial_storage=initial_storage, n_steps=n_steps
                                , n_products=len(products), profiles=profiles, max_storage=max_storage)


def test_can_init_factory(empty_factory):
    assert empty_factory._n_lines == n_lines


def test_schedule_and_job(factory_with_storage):
    profile_ind = 0
    factory = factory_with_storage
    t = 10
    line = factory.profiles[profile_ind].line
    job = Job(profile=profile_ind, time=t, line=line, action='run', contract=None, override=True)
    factory.schedule(job=job)
    assert len(factory._jobs) == 1
    assert factory._jobs[(t, factory.profiles[profile_ind].line)] == job
    assert factory._wallet == initial_wallet
    assert len(factory._storage) == len(initial_storage)
    assert factory._storage == initial_storage

    profile = factory.profiles[profile_ind]

    # nothing happens before job time
    for _ in range(t):
        infos = factory.step()
        assert all(info.is_empty for info in infos)
        assert all(command.action == 'none' for command in factory._commands)

    # the command is executed in job time
    for _ in range(profile.n_steps):
        infos = factory.step()
        assert not any(info.failed for info in infos)
        assert all(command.action == 'none' for i, command in enumerate(factory._commands) if i != line)
        if _ == 0:
            assert isinstance(infos[line].started, RunningCommandInfo)
            assert infos[line].started.beg == factory._next_step - 1
            assert infos[line].started.end == profile.n_steps + factory._next_step - 1
        if _ == profile.n_steps - 1:
            assert isinstance(infos[line].finished, RunningCommandInfo)
        if 0 < _ < profile.n_steps - 1:
            assert isinstance(infos[line].continuing, RunningCommandInfo)
            assert not infos[line].finished and not infos[line].started

    # nothing happens after job time
    for _ in range(t):
        infos = factory.step()
        assert all(info.no_production for info in infos)
        assert all(command.action == 'none' for command in factory._commands)


@mark.parametrize('wallet', [initial_wallet, 0], ids=['has_money', 'no_money'])
def test_schedule_and_job_with_failures(empty_factory, wallet):
    profile_ind = 0
    factory = empty_factory
    factory._wallet = wallet
    t = 10
    line = factory.profiles[profile_ind].line
    job = Job(profile=profile_ind, time=t, line=line, action='run', contract=None, override=True)
    factory.schedule(job=job)
    assert len(factory._jobs) == 1
    assert factory._jobs[(t, factory.profiles[profile_ind].line)] == job
    assert factory._wallet == wallet
    assert len(factory._storage) == 0
    profile = factory.profiles[profile_ind]

    # nothing happens before job time
    for _ in range(t):
        infos = factory.step()
        assert all(info.is_empty for info in infos)
        assert all(command.action == 'none' for command in factory._commands)

    # the command is executed in job time
    infos = factory.step()
    assert not any(info.failed for i, info in enumerate(infos) if i != line)
    assert infos[line].failed
    assert len(infos[line].failure.missing_inputs) > 0
    if wallet < profile.cost:
        assert infos[line].failure.missing_money == profile.cost - wallet

    # nothing happens after job time
    for _ in range(2 * t):
        infos = factory.step()
        assert all(info.no_production for info in infos)
        assert all(command.action == 'none' for command in factory._commands)


def test_slow_factory_simulator_can_be_initialized(slow_simulator):
    assert slow_simulator is not None


def do_simulator_run(simulator, profiles, t, at, profile_ind, override):
    profile = profiles[profile_ind]
    line = profile.line
    cost = profile.cost
    length = profile.n_steps
    total_inputs = sum(_.quantity for _ in profile.process.inputs) if at >= t else 0
    total_outputs = sum(_.quantity for _ in profile.process.outputs) if at >= t + profile.n_steps else 0

    job = Job(profile=profile_ind, time=t, line=line, action='run', contract=None, override=False)
    simulator.schedule(job=job, override=override)
    assert simulator.wallet_at(at) == (initial_wallet if at < t else initial_wallet - cost)
    assert simulator.total_storage_at(at) == sum(initial_storage.values()) - total_inputs + total_outputs
    assert (simulator.line_schedules_at(at)[_] == NO_PRODUCTION for _ in range(n_lines) if _ != line)
    if at < t or at >= t + length:
        assert simulator.line_schedules_at(at)[line] == NO_PRODUCTION

# @todo correct the slow simulator

@mark.parametrize('profile_ind,t,at_,override,simulator_type',
                  [
                      # (0, 10, 'after', True, 'slow'), (0, 10, 'after', False, 'slow'), (0, 10, 'before', True, 'slow'),
                      # (0, 10, 'before', False, 'slow')
                      # , (0, 10, 'just before', True, 'slow'), (0, 10, 'just before', False, 'slow'),
                      # (0, 10, 'just after', True, 'slow')
                      # , (0, 10, 'just after', False, 'slow')
                      # , (0, 10, 'at', True, 'slow'), (0, 10, 'at', False, 'slow'), (0, 10, 'middle', True, 'slow'),
                      # (0, 10, 'middle', False, 'slow')
                      # , (1, 10, 'after', True, 'slow'), (1, 10, 'after', False, 'slow'), (1, 10, 'before', True, 'slow')
                      # , (1, 10, 'before', False, 'slow')
                      # , (1, 10, 'just before', True, 'slow'), (1, 10, 'just before', False, 'slow'),
                      # (1, 10, 'just after', True, 'slow')
                      # , (1, 10, 'just after', False, 'slow')
                      # , (1, 10, 'at', True, 'slow'), (1, 10, 'at', False, 'slow'), (1, 10, 'middle', True, 'slow'),
                      # (1, 10, 'middle', False, 'slow')
                      # , (2, 10, 'after', True, 'slow'), (2, 10, 'after', False, 'slow'), (2, 10, 'before', True, 'slow')
                      # , (2, 10, 'before', False, 'slow')
                      # , (2, 10, 'just before', True, 'slow'), (2, 10, 'just before', False, 'slow'),
                      # (2, 10, 'just after', True, 'slow')
                      # , (2, 10, 'just after', False, 'slow')
                      # , (2, 10, 'at', True, 'slow'), (2, 10, 'at', False, 'slow'), (2, 10, 'middle', True, 'slow'),
                      # (2, 10, 'middle', False, 'slow')
                      # , (3, 10, 'after', True, 'slow'), (3, 10, 'after', False, 'slow'), (3, 10, 'before', True, 'slow')
                      # , (3, 10, 'before', False, 'slow')
                      # , (3, 10, 'just before', True, 'slow'), (3, 10, 'just before', False, 'slow'),
                      # (3, 10, 'just after', True, 'slow')
                      # , (3, 10, 'just after', False, 'slow')
                      # , (3, 10, 'at', True, 'slow'), (3, 10, 'at', False, 'slow'), (3, 10, 'middle', True, 'slow'),
                      # (3, 10, 'middle', False, 'slow'),
                      (0, 0, 'after', True, 'fast'), (0, 0, 'after', False, 'fast'), (0, 0, 'before', True, 'fast'),
                      (0, 0, 'before', False, 'fast')
                      , (0, 0, 'just before', True, 'fast'), (0, 0, 'just before', False, 'fast'),
                      (0, 0, 'just after', True, 'fast')
                      , (0, 0, 'just after', False, 'fast')
                      , (0, 0, 'at', True, 'fast'), (0, 0, 'at', False, 'fast'), (0, 0, 'middle', True, 'fast'),
                      (0, 0, 'middle', False, 'fast')
                      , (1, 0, 'after', True, 'fast'), (1, 0, 'after', False, 'fast'), (1, 0, 'before', True, 'fast')
                      , (1, 0, 'before', False, 'fast')
                      , (1, 0, 'just before', True, 'fast'), (1, 0, 'just before', False, 'fast'),
                      (1, 0, 'just after', True, 'fast')
                      , (1, 0, 'just after', False, 'fast')
                      , (1, 0, 'at', True, 'fast'), (1, 0, 'at', False, 'fast'), (1, 0, 'middle', True, 'fast'),
                      (1, 0, 'middle', False, 'fast')
                      , (2, 0, 'after', True, 'fast'), (2, 0, 'after', False, 'fast'), (2, 0, 'before', True, 'fast')
                      , (2, 0, 'before', False, 'fast')
                      , (2, 0, 'just before', True, 'fast'), (2, 0, 'just before', False, 'fast'),
                      (2, 0, 'just after', True, 'fast')
                      , (2, 0, 'just after', False, 'fast')
                      , (2, 0, 'at', True, 'fast'), (2, 0, 'at', False, 'fast'), (2, 0, 'middle', True, 'fast'),
                      (2, 0, 'middle', False, 'fast')
                      , (3, 0, 'after', True, 'fast'), (3, 0, 'after', False, 'fast'), (3, 0, 'before', True, 'fast')
                      , (3, 0, 'before', False, 'fast')
                      , (3, 0, 'just before', True, 'fast'), (3, 0, 'just before', False, 'fast'),
                      (3, 0, 'just after', True, 'fast')
                      , (3, 0, 'just after', False, 'fast')
                      , (3, 0, 'at', True, 'fast'), (3, 0, 'at', False, 'fast'), (3, 0, 'middle', True, 'fast'),
                      (3, 0, 'middle', False, 'fast')
                      , (0, 10, 'after', True, 'fast'), (0, 10, 'after', False, 'fast'),
                      (0, 10, 'before', True, 'fast'), (0, 10, 'before', False, 'fast')
                      , (0, 10, 'just before', True, 'fast'), (0, 10, 'just before', False, 'fast'),
                      (0, 10, 'just after', True, 'fast')
                      , (0, 10, 'just after', False, 'fast')
                      , (0, 10, 'at', True, 'fast'), (0, 10, 'at', False, 'fast'), (0, 10, 'middle', True, 'fast'),
                      (0, 10, 'middle', False, 'fast')
                      , (1, 10, 'after', True, 'fast'), (1, 10, 'after', False, 'fast'), (1, 10, 'before', True, 'fast')
                      , (1, 10, 'before', False, 'fast')
                      , (1, 10, 'just before', True, 'fast'), (1, 10, 'just before', False, 'fast'),
                      (1, 10, 'just after', True, 'fast')
                      , (1, 10, 'just after', False, 'fast')
                      , (1, 10, 'at', True, 'fast'), (1, 10, 'at', False, 'fast'), (1, 10, 'middle', True, 'fast'),
                      (1, 10, 'middle', False, 'fast')
                      , (2, 10, 'after', True, 'fast'), (2, 10, 'after', False, 'fast'), (2, 10, 'before', True, 'fast')
                      , (2, 10, 'before', False, 'fast')
                      , (2, 10, 'just before', True, 'fast'), (2, 10, 'just before', False, 'fast'),
                      (2, 10, 'just after', True, 'fast')
                      , (2, 10, 'just after', False, 'fast')
                      , (2, 10, 'at', True, 'fast'), (2, 10, 'at', False, 'fast'), (2, 10, 'middle', True, 'fast'),
                      (2, 10, 'middle', False, 'fast')
                      , (3, 10, 'after', True, 'fast'), (3, 10, 'after', False, 'fast'), (3, 10, 'before', True, 'fast')
                      , (3, 10, 'before', False, 'fast')
                      , (3, 10, 'just before', True, 'fast'), (3, 10, 'just before', False, 'fast'),
                      (3, 10, 'just after', True, 'fast')
                      , (3, 10, 'just after', False, 'fast')
                      , (3, 10, 'at', True, 'fast'), (3, 10, 'at', False, 'fast'), (3, 10, 'middle', True, 'fast'),
                      (3, 10, 'middle', False, 'fast')
                      , (0, 0, 'after', True, 'fast'), (0, 0, 'after', False, 'fast'), (0, 0, 'before', True, 'fast'),
                      (0, 0, 'before', False, 'fast')
                      , (0, 0, 'just before', True, 'fast'), (0, 0, 'just before', False, 'fast'),
                      (0, 0, 'just after', True, 'fast')
                      , (0, 0, 'just after', False, 'fast')
                      , (0, 0, 'at', True, 'fast'), (0, 0, 'at', False, 'fast'), (0, 0, 'middle', True, 'fast'),
                      (0, 0, 'middle', False, 'fast')
                      , (1, 0, 'after', True, 'fast'), (1, 0, 'after', False, 'fast'), (1, 0, 'before', True, 'fast')
                      , (1, 0, 'before', False, 'fast')
                      , (1, 0, 'just before', True, 'fast'), (1, 0, 'just before', False, 'fast'),
                      (1, 0, 'just after', True, 'fast')
                      , (1, 0, 'just after', False, 'fast')
                      , (1, 0, 'at', True, 'fast'), (1, 0, 'at', False, 'fast'), (1, 0, 'middle', True, 'fast'),
                      (1, 0, 'middle', False, 'fast')
                      , (2, 0, 'after', True, 'fast'), (2, 0, 'after', False, 'fast'), (2, 0, 'before', True, 'fast')
                      , (2, 0, 'before', False, 'fast')
                      , (2, 0, 'just before', True, 'fast'), (2, 0, 'just before', False, 'fast'),
                      (2, 0, 'just after', True, 'fast')
                      , (2, 0, 'just after', False, 'fast')
                      , (2, 0, 'at', True, 'fast'), (2, 0, 'at', False, 'fast'), (2, 0, 'middle', True, 'fast'),
                      (2, 0, 'middle', False, 'fast')
                      , (3, 0, 'after', True, 'fast'), (3, 0, 'after', False, 'fast'), (3, 0, 'before', True, 'fast')
                      , (3, 0, 'before', False, 'fast')
                      , (3, 0, 'just before', True, 'fast'), (3, 0, 'just before', False, 'fast'),
                      (3, 0, 'just after', True, 'fast')
                      , (3, 0, 'just after', False, 'fast')
                      , (3, 0, 'at', True, 'fast'), (3, 0, 'at', False, 'fast'), (3, 0, 'middle', True, 'fast'),
                      (3, 0, 'middle', False, 'fast')
                  ]
                  )
def test_slow_factory_simulator_with_jobs(products, profiles, profile_ind, t, at_, override, simulator_type):
    simulator_type = SlowFactorySimulator if simulator_type == 'slow' else FastFactorySimulator
    simulator = simulator_type(initial_wallet=initial_wallet, initial_storage=initial_storage, n_steps=n_steps
                               , n_products=len(products), profiles=profiles, max_storage=max_storage)
    profile = profiles[profile_ind]
    length = profile.n_steps
    if at_ == 'before':
        if t - 5 < 0:
            return
        at = t - 5
    elif at_ == 'after':
        if t + length + 3 > n_steps:
            return
        at = t + length + 3
    elif at_ == 'at':
        at = t
    elif at_ == 'just before':
        if t - 1 < 0:
            return
        at = t - 1
    elif at_ == 'just after':
        at = t + length
    elif at_ == 'middle':
        if length == 1:
            return
        at = (t + length) // 2
    else:
        raise ValueError(f'Unknown option {at_}')
    do_simulator_run(simulator, profiles, t, at, profile_ind, override)


if __name__ == '__main__':
    pytest.main(args=[__file__])



from abc import ABC
from collections import defaultdict
from random import random
from typing import TYPE_CHECKING

from dataclasses import dataclass
from numpy.random import dirichlet

from negmas.apps.scml import FinancialReport
from negmas.common import MechanismState, AgentMechanismInterface
from negmas.helpers import ConfigReader, get_class
from negmas.negotiators import Negotiator
from negmas.outcomes import Issue
from negmas.sao import AspirationNegotiator
from negmas.situated import Contract, Breach
from negmas.situated import RenegotiationRequest
from negmas.utilities import LinearUtilityAggregationFunction, normalize
from .agent import SCMLAgent
from .helpers import pos_gauss

if True:
    from typing import Dict, Any, List, Optional
    from .common import Loan

__all__ = [
    'Miner', 'MiningProfile', 'ReactiveMiner'
]


@dataclass
class MiningProfile:
    cv: float = 0.05

    alpha_t: float = 1.0
    alpha_q: float = 1.0
    alpha_u: float = 1.0

    beta_t: float = 1.0
    beta_q: float = 100.0
    beta_u: float = 100.0

    tau_t: float = -0.25
    tau_q: float = 0.25
    tau_u: float = 1.0

    @classmethod
    def random(cls):
        alpha_t, alpha_q, alpha_u = dirichlet((1, 1, 1), size=1)[0]
        tau_t, tau_q, tau_u = 2 * random() - 1, 2 * random() - 1, 2 * random() - 1
        return MiningProfile(cv=random(), alpha_t=alpha_t, alpha_q=alpha_q, alpha_u=alpha_u
                             , tau_t=tau_t, tau_q=tau_q, tau_u=tau_u
                             , beta_t=1.5 * random(), beta_q=99 * random() + 1, beta_u=99 * random() + 1)


class Miner(SCMLAgent, ABC):
    """Base class of all miners"""


class ReactiveMiner(Miner):
    """Raw Material Generator"""

    def on_new_report(self, report: FinancialReport):
        pass

    def on_neg_request_rejected(self, req_id: str, by: Optional[List[str]]):
        pass

    def on_neg_request_accepted(self, req_id: str, mechanism: AgentMechanismInterface):
        pass

    def on_negotiation_success(self, contract: Contract, mechanism: AgentMechanismInterface) -> None:
        pass

    def on_contract_signed(self, contract: Contract) -> None:
        pass

    def on_contract_cancelled(self, contract: Contract, rejectors: List[str]) -> None:
        pass

    def sign_contract(self, contract: Contract) -> Optional[str]:
        return self.id

    def on_contract_nullified(self, contract: Contract, bankrupt_partner: str, compensation: float) -> None:
        pass

    def on_agent_bankrupt(self, agent_id: str) -> None:
        pass

    def confirm_partial_execution(self, contract: Contract, breaches: List[Breach]) -> bool:
        return True

    def on_remove_cfp(self, cfp: 'CFP'):
        pass

    def __init__(self, profiles: Dict[int, MiningProfile] = None, negotiator_type='negmas.sao.AspirationNegotiator'
                 , n_retrials=0, reactive=True, name=None):
        super().__init__(name=name)
        self.negotiator_type = get_class(negotiator_type, scope=globals())
        self.profiles: Dict[int, MiningProfile] = {}
        self.n_neg_trials: Dict[str, int] = defaultdict(int)
        self.n_retrials = n_retrials
        self.reactive = reactive
        if profiles is not None:
            self.set_profiles(profiles=profiles)

    def init(self):
        self.awi.register_interest(list(self.profiles.keys()))

    def on_negotiation_failure(self, partners: List[str], annotation: Dict[str, Any], mechanism: AgentMechanismInterface
                               , state: MechanismState) -> None:
        # noinspection PyUnusedLocal
        cfp = annotation['cfp']
        super().on_negotiation_failure(partners=partners, annotation=annotation, mechanism=mechanism, state=state)
        thiscfp = self.awi.bb_query(section='cfps', query=cfp.id, query_keys=True)
        if cfp.publisher != self.id and thiscfp is not None and len(thiscfp) > 0 \
            and self.n_neg_trials[cfp.id] < self.n_retrials:
            self.awi.logdebug(f'Renegotiating {self.n_neg_trials[cfp.id]} on {cfp}')
            self.on_new_cfp(cfp=annotation['cfp'])

    def set_profiles(self, profiles: Dict[int, MiningProfile]):
        self.profiles = profiles if profiles is not None else dict()

    def _process_cfp(self, cfp: 'CFP'):
        if self.awi.is_bankrupt(cfp.publisher) or not self.can_expect_agreement(cfp=cfp, margin=0):
            return
        profile = self.profiles.get(cfp.product, None)
        if profile is None:
            return
        if profile.cv == 0:
            alpha_u, alpha_q, alpha_t = profile.alpha_u, profile.alpha_q, profile.alpha_t
        else:
            alpha_u, alpha_q, alpha_t = tuple(
                dirichlet((profile.alpha_u, profile.alpha_q, profile.alpha_t), size=1)[0])
        beta_u = pos_gauss(profile.beta_u, profile.cv)
        beta_t = pos_gauss(profile.beta_t, profile.cv)
        beta_q = pos_gauss(profile.beta_q, profile.cv)

        tau_u = pos_gauss(profile.tau_u, profile.cv)
        tau_t = pos_gauss(profile.tau_t, profile.cv)
        tau_q = pos_gauss(profile.tau_q, profile.cv)

        ufun = normalize(LinearUtilityAggregationFunction(issue_utilities={
            'time': lambda x: x ** tau_t / beta_t,
            'quantity': lambda x: x ** tau_q / beta_q,
            'unit_price': lambda x: x ** tau_u / beta_u if x > 1e-7 else -2000.0,
        }, weights={'time': alpha_t, 'quantity': alpha_q, 'unit_price': alpha_u})
            , outcomes=cfp.outcomes, infeasible_cutoff=-1500)
        if self.negotiator_type == AspirationNegotiator:
            negotiator = self.negotiator_type(assume_normalized=True, name=self.name + '*' + cfp.publisher
                                              , dynamic_ufun=False, aspiration_type='boulware')
        else:
            negotiator = self.negotiator_type(assume_normalized=True, name=self.name + '*' + cfp.publisher)
        self.n_neg_trials[cfp.id] += 1
        self.request_negotiation(cfp=cfp, negotiator=negotiator
                                 , ufun=normalize(ufun, outcomes=cfp.outcomes, infeasible_cutoff=None))

    def on_new_cfp(self, cfp: 'CFP'):
        if self.reactive:
            if not cfp.satisfies(query={'products': list(self.profiles.keys())}):
                return
            self._process_cfp(cfp)

    def step(self):
        if not self.reactive:
            cfps = self.awi.bb_query(section='cfps', query_keys=False
                                                 , query={'products': list(self.profiles.keys())})
            if cfps is None:
                return
            cfps = cfps.values()
            for cfp in cfps:
                self._process_cfp(cfp)

    def confirm_contract_execution(self, contract: Contract) -> bool:
        return True

    def respond_to_negotiation_request(self, cfp: "CFP", partner: str) -> Optional[Negotiator]:
        raise ValueError('Miners should never receive negotiation requests as they publish no CFPs')

    def set_renegotiation_agenda(self, contract: Contract, breaches: List[Breach]) -> Optional[RenegotiationRequest]:
        return None

    def respond_to_renegotiation_request(self, contract: Contract, breaches: List[Breach]
                                         , agenda: RenegotiationRequest) -> Optional[Negotiator]:
        return None

    def confirm_loan(self, loan: Loan, bankrupt_if_rejected: bool) -> bool:
        """called by the world manager to confirm a loan if needed by the buyer of a contract that is about to be
        breached"""
        return bankrupt_if_rejected



"""
Tournament generation and management.

"""
import concurrent.futures as futures
import copy
import itertools
import json
import pathlib
import random
import time
import traceback
from os import PathLike
from pathlib import Path
from typing import Optional, List, Callable, Union, Type, Sequence, Dict, Any

import numpy as np
import pandas as pd
import yaml
from dataclasses import dataclass, field
from typing_extensions import Protocol

from .situated import Agent, World, save_stats

from negmas.helpers import get_class, unique_name, import_by_name, get_full_type_name, humanize_time

__all__ = [
    'tournament',
    'WorldGenerator',
    'WorldRunResults',
    'TournamentResults',
    'run_world',
    'process_world_run',
]

PROTOCOL_CLASS_NAME_FIELD = '__mechanism_class_name'

try:
    # disable a warning in yaml 1b1 version
    yaml.warnings({'YAMLLoadWarning': False})
except:
    pass


class WorldGenerator(Protocol):
    """A callback-protocol specifying the signature of a world generator function that can be passed to `tournament`

    Args:
            name: world name. If None, a random name should be generated
            competitors: A list of `Agent` types that can be used to create the agents of the competitor types
            params: A list of parameters to pass to the agent types
            log_File_name: A log file name to keep logs
            randomize: If true, competitors should be assigned randomly within the world. The meaning of "random
            assignment" can vary from a world to another. In general it should be the case that if randomize is False,
            all worlds generated given kwargs, and a selection competitors will be the same.
            agent_names_reveal_type: Whether the type of an agent should be apparent in its name
            kwargs: key-value pairs of arguments.

    See Also:
        `tournament`

    """

    def __call__(self, name: Optional[str] = None, competitors: Sequence[Type[Agent]] = ()
                 , params: Sequence[Dict[str, Any]] = ()
                 , log_file_name: Optional[str] = None, randomize: bool = True, agent_names_reveal_type: bool = False
                 , **kwargs) -> World: ...


@dataclass
class WorldRunResults:
    """Results of a world run"""
    world_name: str
    """World name"""
    log_file_name: str
    """Log file name"""
    names: List[str] = field(default_factory=list, init=False)
    """Agent names"""
    scores: List[float] = field(default_factory=list, init=False)
    """Agent scores"""
    types: List[str] = field(default_factory=list, init=False)
    """Agent type names"""


@dataclass
class TournamentResults:
    scores: pd.DataFrame
    total_scores: pd.DataFrame
    winners: List[str]
    """Winner type name(s) which may be a list"""
    winners_scores: np.array
    """Winner score (accumulated)"""
    ttest: pd.DataFrame


def run_world(world_info: dict):
    """Runs a world and returns stats. This function is designed to be used with distributed systems like dask.

    Args:
        world_info: World info dict. See remarks for its parameters

    Remarks:

        The `world_info` dict should have the following members:

            - name: world name [Defaults to random]
            - competitors: list of strings giving competitor types [Defaults to an empty list]
            - log_file_name: file name to store the world log [Defaults to random]
            - randomize: whether to randomize assignment [Defaults to True]
            - agent_names_reveal_type: whether agent names reveal type [Defaults to False]
            - __dir_name: directory to store the world stats [Defaults to random]
            - __world_generator: full name of the world generator function (including its module) [Required]
            - __score_calculator: full name of the score calculator function [Required]
            - __tournament_name: name of the tournament [Defaults to random]
            - others: values of all other keys are passed to the world generator as kwargs
    """
    world_generator = world_info.get('__world_generator', None)
    score_calculator = world_info.get('__score_calculator', None)
    tournament_name = world_info.get('__tournament_name', unique_name(base=""))
    assert world_generator and score_calculator, f'Cannot run without specifying both a world generator and a score ' \
        f'calculator'

    world_generator = import_by_name(world_generator)
    score_calculator = import_by_name(score_calculator)
    world_info['competitors'] = [get_class(_) for _ in world_info.get('competitors', [])]
    default_name = unique_name(base="")
    world_info['name'] = world_info.get('name', default_name)
    world_name = world_info['name']
    default_dir = (Path(f'~') / 'negmas' / 'tournaments' / tournament_name / world_name).absolute()
    world_info['log_file_name'] = world_info.get('log_file_name', str(default_dir / 'log.txt'))
    world_info['agent_names_reveal_type'] = world_info.get('agent_names_reveal_type', False)
    world_info['randomize'] = world_info.get('randomzie', True)
    world_info['__dir_name'] = world_info.get('__dir_name', str(default_dir))

    # delete the parameters not used by _run_world
    for k in ('__world_generator', '__tournament_name', '__score_calculator'):
        if k in world_info.keys():
            del world_info[k]
    return _run_world(world_info=world_info, world_generator=world_generator, score_calculator=score_calculator)


def _run_world(world_info: dict, world_generator: WorldGenerator,
               score_calculator: Callable[[World], WorldRunResults]
               , world_progress_callback: Callable[[Optional[World]], None] = None
               ):
    """Runs a world and returns stats

    Args:
        world_info: World info dict. See remarks for its parameters
        world_generator: World generator function.
        score_calculator: Score calculator function
        world_progress_callback: world progress callback

    Remarks:

        The `world_info` dict should have the following members:

            - name: world name
            - competitors: list of types giving competitor types
            - log_file_name: file name to store the world log
            - randomize: whether to randomize assignment
            - agent_names_reveal_type: whether agent names reveal type
            - __dir_name: directory to store the world stats
            - others: values of all other keys are passed to the world generator as kwargs
    """
    world_info = world_info.copy()
    dir_name = world_info['__dir_name']
    del world_info['__dir_name']
    world = world_generator(**world_info)
    if world_progress_callback is None:
        world.run()
    else:
        _start_time = time.monotonic()
        for _ in range(world.n_steps):
            if world.time_limit is not None and (time.monotonic() - _start_time) >= world.time_limit:
                break
            if not world.step():
                break
            world_progress_callback(world)
    save_stats(world=world, log_dir=dir_name)
    scores = score_calculator(world)
    return scores, dir_name


def process_world_run(results: WorldRunResults, tournament_name: str, dir_name: str) -> pd.DataFrame:
    """
    Generates a dataframe with the results of this world run

    Args:
        results: Results of the world run
        tournament_name: tournament name
        dir_name: directory name to store the stats.

    Returns:

        A pandas DataFrame with agent_name, agent_type, score, log_file, world, and stats_folder columns

    """
    log_file, world_name_ = results.log_file_name, results.world_name
    if log_file is not None:
        with open(log_file, 'a') as f:
            f.write(f'\nPART of TOURNAMENT {tournament_name}. This world run completed successfully\n')
    scores = []
    for name_, type_, score in zip(results.names, results.types, results.scores):
        scores.append({'agent_name': name_, 'agent_type': type_, 'score': score, 'log_file': log_file
                          , 'world': world_name_, 'stats_folder': dir_name})
    return pd.DataFrame(data=scores)


def _run_dask(scheduler_ip, scheduler_port, verbose, world_infos, world_generator, tournament_progress_callback
              , n_worlds, name, score_calculator) -> List[pd.DataFrame]:
    """Runs the tournament on dask"""

    import distributed
    scores = []
    if scheduler_ip is None and scheduler_port is None:
        address = None
    else:
        if scheduler_ip is None:
            scheduler_ip = '127.0.0.1'
        if scheduler_port is None:
            scheduler_port = '8786'
        address = f'{scheduler_ip}:{scheduler_port}'
    if verbose:
        print(f'Will use DASK on {address}')
    client = distributed.Client(address=address, set_as_default=True)
    future_results = []
    for world_info in world_infos:
        future_results.append(client.submit(_run_world, world_info, world_generator, score_calculator))
    print(f'Submitted all processes to DASK ({len(world_infos)})')
    _strt = time.perf_counter()
    for i, (future, result) in enumerate(
        distributed.as_completed(future_results, with_results=True, raise_errors=False)):
        try:
            score_, dir_name = result
            if tournament_progress_callback is not None:
                tournament_progress_callback(score_, i, n_worlds)
            scores.append(process_world_run(score_, tournament_name=name, dir_name=str(dir_name)))
            if verbose:
                _duration = time.perf_counter() - _strt
                print(f'{i + 1:003} of {n_worlds:003} [{100 * (i + 1) / n_worlds:0.3}%] completed in '
                      f'{humanize_time(_duration)} [ETA {humanize_time(_duration * n_worlds  / (i + 1))}]')
        except Exception as e:
            if tournament_progress_callback is not None:
                tournament_progress_callback(None, i, n_worlds)
            print(traceback.format_exc())
            print(e)
    client.shutdown()
    return scores


def tournament(competitors: Sequence[Union[str, Type[Agent]]]
               , world_generator: WorldGenerator
               , score_calculator: Callable[[World], WorldRunResults]
               , competitor_params: Optional[Sequence[Dict[str, Any]]] = None
               , randomize=False
               , agent_names_reveal_type=False
               , n_agents_per_competitor=1
               , max_n_configs: int = 1000
               , n_runs_per_config: int = 5
               , tournament_path: str = './logs/tournaments'
               , total_timeout: Optional[int] = None
               , parallelism='parallel'
               , scheduler_ip: Optional[str] = None
               , scheduler_port: Optional[str] = None
               , tournament_progress_callback: Callable[[Optional[WorldRunResults], int, int], None] = None
               , world_progress_callback: Callable[[Optional[World]], None] = None
               , name: str = None
               , verbose: bool = False
               , configs_only: bool = False
               , **kwargs
               ) -> Union[TournamentResults, PathLike]:
    """
    Runs a tournament

    Args:

        n_agents_per_competitor: The number of agents of each competing type to be instantiated in the world
        name: Tournament name
        world_generator: A functions to generate worlds for the tournament
        score_calculator: A function for calculating the score of a world *After it finishes running*
        competitors: A list of class names for the competitors
        competitor_params: A list of competitor parameters (used to initialize the competitors).
        randomize: If true, then instead of trying all possible permutations of assignment random shuffles will be used.
        agent_names_reveal_type: If true then the type of an agent should be readable in its name (most likely at its
        beginning).
        max_n_configs: No more than n_runs_max worlds will be run. If `randomize` then it cannot be None and that is exactly
        the number of worlds to run. If not `randomize` then at most this number of worlds will be run if it is not None
        n_runs_per_config: Number of runs per configuration.
        total_timeout: Total timeout for the complete process
        tournament_path: Path at which to store all results. A scores.csv file will keep the scores and logs folder will
        keep detailed logs
        parallelism: Type of parallelism. Can be 'serial' for serial, 'parallel' for parallel and 'distributed' for distributed
        scheduler_port: Port of the dask scheduler if parallelism is dask, dist, or distributed
        scheduler_ip:   IP Address of the dask scheduler if parallelism is dask, dist, or distributed
        world_progress_callback: A function to be called after everystep of every world run (only allowed for serial
        evaluation and should be used with cautious).
        tournament_progress_callback: A function to be called with `WorldRunResults` after each world finished
        processing
        verbose: Verbosity
        configs_only: If true, a config file for each
        kwargs: Arguments to pass to the `world_generator` function

    Returns:
        `TournamentResults` The results of the tournament or a `PathLike` giving the location where configs were saved

    """
    dask_options = ('dist', 'distributed', 'dask', 'd')
    multiprocessing_options = ('local', 'parallel', 'par', 'p')
    serial_options = ('none', 'serial', 's')
    if parallelism is None:
        parallelism = 'serial'
    assert total_timeout is None or parallelism not in dask_options, f'Cannot use {parallelism} with a total-timeout'
    assert world_progress_callback is None or parallelism not in dask_options, f'Cannot use {parallelism} with a world callback'
    if name is None:
        name = unique_name('', add_time=True, rand_digits=0)
    competitors = list(competitors)
    if tournament_path.startswith('~'):
        tournament_path = Path.home() / ('/'.join(tournament_path.split('/')[1:]))
    tournament_path = (pathlib.Path(tournament_path) / name).absolute()
    tournament_path.mkdir(parents=True, exist_ok=True)
    if verbose:
        print(f'Results of Tournament {name} will be saved to {str(tournament_path)}')
    if competitor_params is None:
        competitor_params = [dict() for _ in range(len(competitors))]
    params = {
        'competitors': [get_class(_).__name__ if not isinstance(_, str) else _ for _ in competitors],
        'params': competitor_params,
        'randomize': randomize,
        'n_runs': max_n_configs,
        'tournament_path': str(tournament_path),
        'total_timeout': total_timeout,
        'parallelism': parallelism,
        'scheduler_ip': scheduler_ip,
        'scheduler_port': scheduler_port,
        'name': name,
        'n_worlds_to_run': None
    }
    params.update(kwargs)
    with (tournament_path / 'params.json').open('w') as f:
        json.dump(params, f, sort_keys=True, indent=4)
    world_infos = []
    if randomize:
        for i in range(max_n_configs):
            competitor_infos = list(zip(competitors, competitor_params))
            random.shuffle(competitor_infos)
            competitors = [_[0] for _ in competitor_infos]
            competitor_params = [_[1] for _ in competitor_infos]
            world_name = unique_name(f'{i:05}', add_time=True, rand_digits=4)
            dir_name = tournament_path / world_name
            world_info = {'name': world_name, 'competitors': competitors, 'params': competitor_params
                , 'log_file_name': str(dir_name / 'log.txt')
                , 'randomize': True, 'agent_names_reveal_type': agent_names_reveal_type
                          , 'n_agents_per_competitor': n_agents_per_competitor
                , '__dir_name': str(dir_name)}
            world_info.update(kwargs)
            world_infos += [world_info.copy() for _ in range(n_runs_per_config)]
    else:
        competitor_infos = list(zip(competitors, competitor_params))
        c_list_infos = list(itertools.permutations(competitor_infos))
        if max_n_configs is not None:
            if len(c_list_infos) > max_n_configs:
                print(f'Need {len(c_list_infos)} permutations but allowed to only use {max_n_configs} of them'
                      f' ({max_n_configs / len(c_list_infos):0.2%})')
                random.shuffle(c_list_infos)
                c_list_infos = c_list_infos[:max_n_configs]

        for i, cs in enumerate(c_list_infos):
            c = [_[0] for _ in cs]
            ps = [_[1] for _ in cs]
            world_name = unique_name(f'{i:05}', add_time=True, rand_digits=4)
            dir_name = tournament_path / world_name
            world_info = {'name': world_name, 'competitors': list(c), 'params': list(ps)
                , 'log_file_name': str(dir_name / 'log.txt')
                , 'randomize': False, 'agent_names_reveal_type': agent_names_reveal_type
                , 'n_agents_per_competitor': n_agents_per_competitor
                , '__dir_name': str(dir_name)}
            world_info.update(kwargs)
            world_infos += [world_info.copy() for _ in range(n_runs_per_config)]

    saved_configs = [{k: copy.copy(v) if k != 'competitors' else
    [get_full_type_name(c) if not isinstance(c, str) else c for c in v]
                      for k, v in _.items()} for _ in world_infos]
    score_calculator_name = get_full_type_name(score_calculator) if not isinstance(score_calculator,
                                                                                   str) else score_calculator
    world_generator_name = get_full_type_name(world_generator) if not isinstance(world_generator,
                                                                                 str) else world_generator
    for d in saved_configs:
        d['__score_calculator'] = score_calculator_name
        d['__world_generator'] = world_generator_name
        d['__tournament_name'] = name
    config_path = tournament_path / 'configs'
    config_path.mkdir(exist_ok=True, parents=True)
    for i, conf in enumerate(saved_configs):
        f_name = config_path / f'{i:06}.json'
        with open(f_name, 'w') as f:
            json.dump(conf, f, sort_keys=True, indent=4)

    if configs_only:
        return config_path

    scores = []
    scores_file = str(tournament_path / 'scores.csv')
    n_worlds = len(world_infos)
    params['n_worlds_to_run'] = n_worlds
    with (tournament_path / 'params.json').open('w') as f:
        json.dump(params, f, sort_keys=True, indent=4)
    if verbose:
        print(f'Will run {n_worlds} worlds')
    if parallelism in serial_options:
        strt = time.perf_counter()
        for i, world_info in enumerate(world_infos):
            if total_timeout is not None and time.perf_counter() - strt > total_timeout:
                break
            try:
                score_, _ = _run_world(world_info=world_info, world_generator=world_generator
                                       , world_progress_callback=world_progress_callback
                                       , score_calculator=score_calculator)
                if tournament_progress_callback is not None:
                    tournament_progress_callback(score_, i, n_worlds)
                scores.append(process_world_run(score_, tournament_name=name, dir_name=str(world_info['__dir_name'])))
                if verbose:
                    _duration = time.perf_counter() - strt
                    print(f'{i + 1:003} of {n_worlds:003} [{100 * (i + 1) / n_worlds:0.3}%] completed '
                          f'in {humanize_time(_duration)}'
                          f' [ETA {humanize_time(_duration * n_worlds  / (i + 1))}]')
            except Exception as e:
                if tournament_progress_callback is not None:
                    tournament_progress_callback(None, i, n_worlds)
                print(traceback.format_exc())
                print(e)
    elif parallelism in multiprocessing_options:
        executor = futures.ProcessPoolExecutor(max_workers=None)
        future_results = []
        for world_info in world_infos:
            future_results.append(executor.submit(_run_world, world_info, world_generator, score_calculator
                                                  , world_progress_callback))
        if verbose:
            print(f'Submitted all processes ({len(world_infos)})')
        _strt = time.perf_counter()
        for i, future in enumerate(futures.as_completed(future_results, timeout=total_timeout)):
            try:
                score_, dir_name = future.result()
                if tournament_progress_callback is not None:
                    tournament_progress_callback(score_, i, n_worlds)
                scores.append(process_world_run(score_, tournament_name=name, dir_name=str(dir_name)))
                if verbose:
                    _duration = time.perf_counter() - _strt
                    print(f'{i+1:003} of {n_worlds:003} [{100*(i+1)/n_worlds:0.3}%] completed in '
                          f'{humanize_time(_duration)}'
                          f' [ETA {humanize_time(_duration * n_worlds  / (i + 1))}]')
            except futures.TimeoutError:
                if tournament_progress_callback is not None:
                    tournament_progress_callback(None, i, n_worlds)
                print('Tournament timed-out')
                break
            except Exception as e:
                if tournament_progress_callback is not None:
                    tournament_progress_callback(None, i, n_worlds)
                print(traceback.format_exc())
                print(e)
    elif parallelism in dask_options:
        scores = _run_dask(scheduler_ip, scheduler_port, verbose, world_infos, world_generator
                           , tournament_progress_callback, n_worlds, name, score_calculator)
    if verbose:
        print(f'Finding winners')
    if len(scores) < 1:
        return TournamentResults(scores=pd.DataFrame(), total_scores=pd.DataFrame()
                          , winners=[], winners_scores=np.array([])
                          , ttest=pd.DataFrame())
    scores: pd.DataFrame = pd.concat(scores, ignore_index=True)
    scores = pd.DataFrame(data=scores)
    scores.to_csv(scores_file, index_label='index')
    scores = scores.loc[~scores['agent_type'].isnull(), :]
    scores = scores.loc[scores['agent_type'].str.len() > 0, :]
    total_scores = scores.groupby(['agent_type'])['score'].mean().sort_values(ascending=False).reset_index()
    winner_table = total_scores.loc[total_scores['score'] == total_scores['score'].max(), :]
    winners = winner_table['agent_type'].values.tolist()
    winner_scores = winner_table['score'].values
    types = list(scores['agent_type'].unique())

    ttest_results = []
    for i, t1 in enumerate(types):
        for j, t2 in enumerate(types[i + 1:]):
            from scipy.stats import ttest_ind
            t, p = ttest_ind(scores[scores['agent_type'] == t1].score, scores[scores['agent_type'] == t2].score)
            ttest_results.append({'a': t1, 'b': t2, 't': t, 'p': p})

    if verbose:
        print(f'Tournament completed successfully\nWinners: {list(zip(winners, winner_scores))}')

    scores.to_csv(str(tournament_path / 'scores.csv'), index_label='index')
    total_scores.to_csv(str(tournament_path / 'total_scores.csv'), index_label='index')
    winner_table.to_csv(str(tournament_path / 'winners.csv'), index_label='index')
    ttest_results = pd.DataFrame(data=ttest_results)
    ttest_results.to_csv(str(tournament_path / 'ttest.csv'), index_label='index')

    return TournamentResults(scores=scores, total_scores=total_scores, winners=winners, winners_scores=winner_scores
                             , ttest=ttest_results)



"""Common data-structures and objects used throughout the SCM world implementation"""
import itertools
import math
import sys
import uuid
from collections import defaultdict, namedtuple
from dataclasses import dataclass, field, InitVar
from typing import Dict, Union, Tuple, Iterable, List, Optional, Any

import numpy as np

from negmas.outcomes import OutcomeType, Issue
from negmas.situated import Contract

INVALID_STEP = -1000
NO_PRODUCTION = -1
INVALID_UTILITY = -2000

g_last_product_id = 0
g_last_process_id = 0

__all__ = ['Product', 'Process', 'InputOutput', 'RunningCommandInfo'
    , 'INVALID_STEP', 'NO_PRODUCTION', 'INVALID_UTILITY'
    , 'ManufacturingProfile', 'ManufacturingProfileCompiled', 'ProductManufacturingInfo'
    , 'FactoryStatusUpdate', 'Job', 'ProductionNeed'
    , 'MissingInput', 'ProductionReport', 'ProductionFailure', 'FinancialReport'
    , 'SCMLAgreement', 'SCMLAction'
    , 'CFP', 'Loan', 'InsurancePolicy', 'Factory']


@dataclass
class Product:
    __slots__ = ['id', 'production_level', 'name', 'expires_in', 'catalog_price']
    """A product that can be transacted in."""
    id: int
    """Product index. Must be set during construction and **MUST** be unique for products in the same world"""
    production_level: int
    """The level of this product in the production graph."""
    name: str
    """Object name"""
    expires_in: Optional[int]
    """Number of steps within which the product must be consumed. None means never"""
    catalog_price: Optional[float]
    """Catalog price of the product."""

    def __str__(self):
        """String representation is simply the name"""
        return self.name + (f'(cp:{self.catalog_price})' if self.catalog_price is not None else "")

    def __post_init__(self):
        global g_last_product_id
        if self.id is None:
            self.id = g_last_product_id
            g_last_product_id += 1
        if self.name is None:
            self.name = str(self.id)

    def __hash__(self):
        return self.id.__hash__()


@dataclass(frozen=True)
class InputOutput:
    """An input/output to a production process"""
    __slots__ = ['product', 'quantity', 'step']
    product: int
    """Index of the product used as input or output"""
    quantity: int
    """Quantity needed/produced"""
    step: float
    """Relative time within the production at which the input is needed (output is produced)"""


@dataclass
class Process:
    __slots__ = ['id', 'production_level', 'name', 'inputs', 'outputs', 'historical_cost']
    id: int
    """A manufacturing process."""
    production_level: int
    """The level of this process in the production graph"""
    name: str
    """Object name"""
    inputs: List[InputOutput]
    """list of input product name + quantity required and time of consumption relative to the time required for 
    production (value from 0 to 1)"""
    outputs: List[InputOutput]
    """list of output product names, quantity required and when it becomes available relative to the time required for 
    production (value from 0 to 1)"""
    historical_cost: Optional[float]
    """Average cost for running this process in some world. Filled by the world"""

    def __str__(self):
        """String representation is simply the name"""
        return self.name + (f'(cp:{self.historical_cost})' if self.historical_cost is not None else "")

    def __post_init__(self):
        global g_last_process_id
        if self.id is None:
            self.id = g_last_process_id
            g_last_process_id += 1
        if self.name is None:
            self.name = str(self.id)

    def __hash__(self):
        """The hash depends only on the name"""
        return str(self).__hash__()


@dataclass
class ManufacturingProfile:
    """The costs/time required for running a process on a line (with associated cancellation costs etc). This
    data-structure carries full information about the `Process` es instead of just its index as in
    `ManufacturingProfileCompiled`. It is intended to be used to construct factories

    See Also:
        `Factory`

    """
    __slots__ = ['n_steps', 'cost', 'initial_pause_cost', 'running_pause_cost', 'resumption_cost'
        , 'cancellation_cost', 'line', 'process']
    n_steps: int
    """Number of steps needed to complete the manufacturing"""
    cost: float
    """Cost of manufacturing"""
    initial_pause_cost: float
    """Cost of pausing incurred only at the step a pause is started"""
    running_pause_cost: float
    """Running cost of pausing"""
    resumption_cost: float
    """Cost of resuming a process"""
    cancellation_cost: float
    """Cost of cancelling the process before the last step"""
    line: int
    """The line index"""
    process: Process
    """The `Process` associated with this profile"""


@dataclass
class FactoryStatusUpdate:
    __slots__ = ['balance', 'storage']
    balance: float
    """The update to the balance"""
    storage: Dict[int, int]
    """The updates to be applied to the storage after this step"""

    def __post_init__(self):
        if not isinstance(self.storage, defaultdict):
            self.storage = defaultdict(int, self.storage)

    def make_empty(self) -> None:
        """Makes the update an empty one."""
        self.balance = 0.0
        self.storage = defaultdict(int)

    def combine(self, other: "FactoryStatusUpdate") -> None:
        """
        Combines this status update with another one in place

        Args:
            other: The other status update

        Returns:
            None
        """
        if other is None:
            return
        self.balance += other.balance
        to_remove = []
        for k in itertools.chain(self.storage.keys(), other.storage.keys()):
            self.storage[k] += other.storage[k]
            if self.storage[k] == 0:
                to_remove.append(k)
        for k in to_remove:
            del self.storage[k]

    @classmethod
    def combine_sets(cls, dst: Dict[int, "FactoryStatusUpdate"], src: Dict[int, "FactoryStatusUpdate"]):
        """
        Combines a set of updates over time with another in place (overriding `first`)
        Args:
            dst: First set of updates to be combined into
            src: second set of updates to be combined from

        Returns:

        """
        to_remove = []
        for i, update in src.items():
            dst[i].combine(update)
            if dst[i].balance == 0 and len(dst[i].storage) == 0:
                to_remove.append(i)
        for i in to_remove:
            del dst[i]
        return None

    @property
    def is_empty(self):
        return self.balance == 0 and (len(self.storage) == 0 or sum(self.storage.values()) == 0)

    @classmethod
    def empty(cls):
        return FactoryStatusUpdate(balance=0.0, storage={})

    def __str__(self):
        return f'balance: {self.balance}, ' + \
               f'{str({k: v for k, v in self.storage.items()}) if self.storage is not None else ""}'


@dataclass
class RunningCommandInfo:
    __slots__ = ['profile', 'beg', 'end', 'action', 'updates', 'step', 'paused']
    profile: ManufacturingProfile
    """The manufacturing profile associated with this command. Most importantly, it gives the process and line"""
    beg: int
    """The time the command is to be executed"""
    end: int
    """The number of steps starting at `beg` for this command to end (it ends at end - 1)"""
    step: int
    """The time-step relative to `beg` at the factory is currently executing the `Process` indicated in `profile`.
    `step` will always go up by one every simulation step except if the command is paused where it does not change
    """
    paused: bool
    """True if the command is paused"""
    action: str
    """The command type. For the current implementation it will always be run or none for no command"""
    updates: Dict[int, 'FactoryStatusUpdate']
    """The status updates implied by this command with their times relative to `beg`"""

    @property
    def n_steps(self) -> int:
        return self.end - self.beg

    def ended_before(self, t: int):
        return self.end <= t

    def started_on_or_after(self, t: int):
        return self.beg >= t

    def __str__(self):
        # if self.is_none:
        #     return 'No command'
        return f'{self.action} {self.profile.process.id} [{self.beg}:{self.end - 1}]'

    @property
    def is_none(self):
        return self.action == 'none'

    @is_none.setter
    def is_none(self, is_none):
        self.action = 'none'

    @classmethod
    def do_nothing(cls):
        # noinspection PyTypeChecker
        return cls(profile=None, beg=-1, end=-1, action='none', updates={}, step=0, paused=False)


@dataclass
class Job:
    """Describes a job to be run on one production line of a `Factory`."""
    __slots__ = ['profile', 'time', 'line', 'action', 'contract', 'override']
    profile: int
    """The process for run commands"""
    time: int
    """The time the command is to be executed"""
    line: int
    """Index of the line on which the job is to be scheduled. Notice that it will be ignored for `run` actions."""
    action: str
    """The command type. For the current implementation it can be run/pause/resume/stop/cancel with `cancel` cancelling 
    any other command type."""
    contract: Optional[Contract]
    """The sell contract associated with the command"""
    override: bool
    """Whether to override existing commands when the job is to be executed."""

    def __str__(self):
        s = f'{self.action} {self.profile if self.action == "run" else ""} at {self.time} on {self.line}'
        s += f'{" override" if self.override else ""}'
        if self.contract is not None:
            s += f' for {self.contract.id}'
        return s

    def is_cancelling(self, job: 'Job') -> bool:
        """
        Determines if the given jobs cancels this one

        Args:
            job:

        Returns:

        """
        if self.line != job.line:
            return False
        return job.action == 'cancel' or \
               (self.action in ('run', 'start') and job.action == 'stop') or \
               (self.action == 'pause' and job.action == 'resume') or \
               (self.action == 'resume' and job.action == 'pause')


@dataclass
class ProductionNeed:
    """Describes some quantity of a product that is needed to honor a (sell) contract."""
    __slots__ = ['product', 'needed_for', 'quantity_to_buy', 'quantity_in_storage', 'step']
    product: int
    """The product needed"""
    needed_for: Contract
    """The contract for which the product is needed"""
    quantity_to_buy: int
    """The quantity need to be bought"""
    quantity_in_storage: int
    """The quantity already found in storage"""
    step: int
    """The time step at which the product is needed"""

    def __str__(self):
        """String representation is simply the name"""
        return f'Need {self.quantity_to_buy} ({self.quantity_in_storage} exist) of {self.product} at ' + \
               f' {self.step} for {self.needed_for}'


@dataclass
class MissingInput:
    __slots__ = ['product', 'quantity']
    product: int
    quantity: int

    def __str__(self):
        return f'{self.product}: {self.quantity}'


@dataclass
class ProductionFailure:
    __slots__ = ['line', 'command', 'missing_inputs', 'missing_money', 'missing_space']
    line: int
    """ID of the line that failed"""
    command: RunningCommandInfo
    """Information about the command that failed"""
    missing_inputs: List[MissingInput]
    """The missing inputs if any with their quantities"""
    missing_money: float
    """The amount of money needed for production that is not available"""
    missing_space: int
    """The amount space needed in storage but not found"""

    def __str__(self):
        s = f'{str(self.command)} on {self.line} failed:'
        if self.missing_money > 0:
            s += f' money {self.missing_money}'
        if len(self.missing_inputs) > 0:
            s += f' inputs: {[str(_) for _ in self.missing_inputs]}'
        if self.missing_space > 0:
            s += f' space {self.missing_space}'
        return s


@dataclass
class ProductionReport:
    line: int
    """ID of the line"""
    started: Optional[RunningCommandInfo]
    """Commands started"""
    continuing: Optional[RunningCommandInfo]
    """Command that is continuing"""
    finished: Optional[RunningCommandInfo]
    """Command finished"""
    failure: Optional[ProductionFailure]
    """Failures"""
    updates: FactoryStatusUpdate
    """Updates applied to the factory"""

    @property
    def failed(self):
        return self.failure is not None

    @property
    def is_empty(self):
        return self.no_production and self.updates.is_empty

    @property
    def no_production(self):
        return self.started is None and self.finished is None and self.failure is None

    def __str__(self):
        if self.is_empty:
            return ''
        s = f'{self.line}: '
        if self.failed:
            s += f'{str(self.failure)} '
        else:
            if self.started is not None and self.finished is not None:
                s += f'started/finished {str(self.started)} '
            elif self.started is not None:
                s += f'started {str(self.started)} '
            elif self.finished is not None:
                s += f'finished {str(self.finished)} '
        if not self.updates.is_empty:
            s += f'{str(self.updates)}'
        return s


@dataclass
class SCMLAgreement(OutcomeType):
    time: int
    """delivery time"""
    unit_price: float
    """unit price"""
    quantity: int
    """quantity"""
    penalty: Optional[float] = None
    """penalty"""
    signing_delay: int = -1
    """Delay between agreement conclusion and signing it to be binding"""


@dataclass
class CFP(OutcomeType):
    """A Call for proposal upon which a negotiation can start"""
    is_buy: bool
    """If true, the author wants to buy otherwise to sell. Non-negotiable."""
    publisher: str
    """the publisher name. Non-negotiable."""
    product: int
    """product ID. Non-negotiable."""
    time: Union[int, Tuple[int, int], List[int]]
    """delivery time. May be negotiable."""
    unit_price: Union[float, Tuple[float, float], List[float]]
    """unit price. May be negotiable."""
    quantity: Union[int, Tuple[int, int], List[int]]
    """quantity. May be negotiable."""
    penalty: Optional[Union[float, Tuple[float, float], List[float]]] = None
    """penalty per missing item in case the seller cannot provide the required quantity. May be negotiable."""
    signing_delay: Optional[Union[int, Tuple[int, int], List[int]]] = None
    """The grace period after which the agents are asked to confirm signing the contract"""
    money_resolution: Optional[float] = None
    """If not None then it is the minimum unit of money (e.g. 1 for dollar, 0.01 for cent, etc)"""
    id: str = field(default_factory=lambda: str(uuid.uuid4()), init=True)
    """Unique CFP ID"""

    def __str__(self):
        s = f'{self.publisher}: {"buy" if self.is_buy else "sell"} '
        s += f'{self.product} '
        s += f'(t: {self.time}, u: {self.unit_price}, q: {self.quantity}'
        if self.penalty is not None:
            s += f', penalty: {self.penalty}'
        if self.signing_delay is not None:
            s += f', sign after: {self.signing_delay}'
        s += ')'
        return s

    def satisfies(self, query: Dict[str, Any]) -> bool:
        """
        Tests whether the CFP satisfies the conditions set by the query

        Args:

            query: A dictionary given the conditions. See `Remarks` for details

        Remarks:

           - The query dictionary can be used to specify any conditions that are required in the CFP. Only CFPs that
             satisfy ALL the conditions specified in the query are considered satisfying the query. The following keys
             can be set with corresponding meanings:

             is_buy
               True or False. If both are OK, just do not add this key

             publisher
               A string or `SCMLAgent` specifying a specific publisher

             publishers
               A list of publishers (see publisher key)

             product
              A string specifying a product name

             products
               A list of products (see product key)

             time
               A number, list or 2-items-tuple (range) specifying possible times to consider satisfactory

             unit_price
               A number, list or 2-items-tuple (range) specifying possible prices to consider satisfactory

             quantity
                A number, list or 2-items-tuple (range) specifying possible quantities to consider OK

             penalty
               A number, list or 2-items-tuple (range) specifying possible penalties to consider satisfactory

        """

        def _overlap(a: Union[int, float, Tuple[float, float], List[float], Tuple[int, int], List[int]]
                     , b: Union[float, Tuple[float, float], List[float], int, Tuple[int, int], List[int]]):

            def _test_single(a, b):
                if not isinstance(b, Iterable):
                    return a == b
                if isinstance(b, tuple):
                    return b[0] <= a <= b[1]
                return a in b

            if not isinstance(b, Iterable):
                a, b = b, a

            if not isinstance(a, Iterable):
                return _test_single(a, b)

            if isinstance(a, tuple):
                if isinstance(b, tuple):
                    return b[0] <= a[0] <= b[1] or b[0] <= a[1] <= b[1]
                return any(_test_single(_, a) for _ in b)  # type: ignore

            return any(_test_single(_, b) for _ in a)

        for k, v in query.items():
            if k == 'is_buy' and self.is_buy != v:
                return False
            if k == 'publisher' and self.publisher != v:
                return False
            if k == 'publishers' and self.publisher not in v:
                return False
            if k == 'products' and self.product not in v:
                return False
            if k == 'product_ids' and self.product not in v:
                return False
            if k == 'product_indices' and self.product not in v:
                return False
            if k == 'product' and self.product != v:
                return False
            if k == 'product_id' and self.product != v:
                return False
            if k == 'product_index' and self.product != v:
                return False
            if k == 'time' and not _overlap(v, self.time):
                return False
            if k == 'unit_price' and not _overlap(v, self.unit_price):
                return False
            if k == 'penalty':
                if self.penalty is None and v is None:
                    return True
                if self.penalty is None or v is None:
                    return False
                if not _overlap(v, self.penalty):
                    return False
            if k == 'quantity' and not _overlap(v, self.quantity):
                return False
        return True

    @property
    def issues(self):
        """Returns the set of issues associated with this CFP. Notice that some of the issues may have a single value"""

        def _values(x, ensure_list=False, ensure_int=False):
            if isinstance(x, tuple) and ensure_list:
                if x[0] == x[1]:
                    if ensure_list and self.money_resolution is not None:
                        if ensure_int:
                            return [int(math.floor(x[0] / self.money_resolution) * self.money_resolution)]
                        return [math.floor(x[0] / self.money_resolution) * self.money_resolution]
                    else:
                        if ensure_int:
                            return [int(x[0])]
                        return [x[0]]
                if isinstance(x[0], float) or isinstance(x[1], float):
                    xs = (int(math.floor(x[0] / self.money_resolution))
                          , int(math.floor(x[1] / self.money_resolution)))
                    xs = list(_ * self.money_resolution for _ in range(xs[0], xs[1] + 1))
                elif isinstance(x[0], int):
                    xs = list(range(x[0], x[1] + 1))
                else:
                    xs = list(range(int(x[0]), int(x[1]) + 1))
                if len(xs) == 0:
                    if ensure_list and self.money_resolution is not None:
                        if ensure_int:
                            return [int(math.floor(x[0] / self.money_resolution) * self.money_resolution)]
                        return [math.floor(x[0] / self.money_resolution) * self.money_resolution]
                    if ensure_int:
                        return [int(x[0])]
                    return [x[0]]
                if ensure_int:
                    return list(set(int(_) for _ in xs))
                return list(set(xs))
            if isinstance(x, Iterable):
                if ensure_int:
                    return list(set(int(_) for _ in x))
                return list(set(x))
            if ensure_int:
                return [int(x)]
            return [x]

        issues = [Issue(name='time', values=_values(self.time, ensure_list=True, ensure_int=True))
            , Issue(name='quantity', values=_values(self.quantity, ensure_list=True, ensure_int=True))
            , Issue(name='unit_price', values=_values(self.unit_price, ensure_list=self.money_resolution is not None))]
        if self.penalty is not None:
            issues.append(Issue(name='penalty'
                                , values=_values(self.penalty, ensure_list=self.money_resolution is not None)))
        if self.signing_delay is not None:
            issues.append(Issue(name='signing_delay', values=_values(self.quantity, ensure_list=True, ensure_int=True)))
        return issues

    @property
    def outcomes(self):
        return Issue.enumerate(issues=self.issues, max_n_outcomes=1000)

    @property
    def min_time(self):
        if isinstance(self.time, tuple):
            return self.time[0]
        elif isinstance(self.time, list):
            return min(self.time)
        return self.time

    @property
    def max_time(self):
        if isinstance(self.time, tuple):
            return self.time[1]
        elif isinstance(self.time, list):
            return max(self.time)
        return self.time

    @property
    def min_quantity(self):
        if isinstance(self.quantity, tuple):
            return self.quantity[0]
        elif isinstance(self.quantity, list):
            return min(self.quantity)
        return self.quantity

    @property
    def max_quantity(self):
        if isinstance(self.quantity, tuple):
            return self.quantity[1]
        elif isinstance(self.quantity, list):
            return max(self.quantity)
        return self.quantity

    @property
    def min_unit_price(self):
        if isinstance(self.unit_price, tuple):
            return self.unit_price[0]
        elif isinstance(self.unit_price, list):
            return min(self.unit_price)
        return self.unit_price

    @property
    def max_unit_price(self):
        if isinstance(self.unit_price, tuple):
            return self.unit_price[1]
        elif isinstance(self.unit_price, list):
            return max(self.unit_price)
        return self.unit_price

    @property
    def min_signing_delay(self):
        if self.signing_delay is None:
            return None
        if isinstance(self.signing_delay, tuple):
            return self.signing_delay[0]
        elif isinstance(self.signing_delay, list):
            return min(self.signing_delay)
        return self.signing_delay

    @property
    def max_signing_delay(self):
        if self.signing_delay is None:
            return None
        if isinstance(self.signing_delay, tuple):
            return self.signing_delay[1]
        elif isinstance(self.signing_delay, list):
            return max(self.signing_delay)
        return self.signing_delay

    @property
    def min_penalty(self):
        if self.penalty is None:
            return None
        if isinstance(self.penalty, tuple):
            return self.penalty[0]
        elif isinstance(self.penalty, list):
            return min(self.penalty)
        return self.penalty

    @property
    def max_penalty(self):
        if self.penalty is None:
            return None
        if isinstance(self.penalty, tuple):
            return self.penalty[1]
        elif isinstance(self.penalty, list):
            return max(self.penalty)
        return self.penalty

    def to_java(self):
        d = self.__dict__
        d['min_time'], d['max_time'] = int(self.min_time), int(self.max_time)
        d['min_quantity'], d['max_quantity'] = int(self.min_quantity), int(self.max_quantity)
        d['min_unit_price'], d['max_unit_price'] = float(self.min_unit_price), float(self.max_unit_price)
        d['min_penalty'] = float(self.min_penalty) if self.min_penalty is not None else None
        d['max_penalty'] = float(self.max_penalty) if self.max_penalty is not None else None
        d['min_signing_delay'] = int(self.min_signing_delay) if self.min_signing_delay is not None else None
        d['max_signing_delay'] = int(self.max_signing_delay) if self.max_signing_delay is not None else None
        d['money_resolution'] = float(self.money_resolution) if self.money_resolution is not None else 0.0
        return d

    @classmethod
    def from_java(cls, idict: Dict[str, Any]) -> 'CFP':
        if idict['min_time'] == idict['max_time']:
            t = idict['min_time']
        else:
            t = (idict['min_time'], idict['max_time'])
        if idict['min_quantity'] == idict['max_quantity']:
            q = idict['min_quantity']
        else:
            q = (idict['min_quantity'], idict['max_quantity'])
        if idict['min_unit_price'] == idict['max_unit_price']:
            up = idict['min_unit_price']
        else:
            up = (idict['min_unit_price'], idict['max_unit_price'])
        if not idict.get('min_penalty', None) or not idict.get('max_penalty', None):
            p = None
        else:
            if idict['min_penalty'] == idict['max_penalty']:
                p = idict['min_penalty']
            else:
                p = (idict['min_penalty'], idict['max_penalty'])
        if not idict.get('min_signing_delay', None) or not idict.get('max_signing_delay', None):
            s = None
        else:
            if idict['min_signing_delay'] == idict['max_signing_delay']:
                s = idict['min_signing_delay']
            else:
                s = (idict['min_signing_delay'], idict['max_signing_delay'])

        return cls(is_buy=idict['is_buy'], publisher=idict['publisher'], product=idict['product'], time=t, unit_price=up
                   , quantity=q, penalty=p, signing_delay=s, money_resolution=idict.get('money_resolution', None)
                   , id=idict.get('id', None))


@dataclass
class SCMLAction:
    line: str
    """Line to execute the action on (need not be given if the profile is given"""
    profile: Optional[int]
    """Index of the profile to execute"""
    action: str
    """The action which may be start, stop, pause, resume"""
    time: int = 0
    """Time to execute the action at"""


@dataclass
class ManufacturingProfileCompiled:
    """The costs/time required for running a process on a line (with associated cancellation costs etc).

    See Also:
        `Factory`

    """
    __slots__ = ['n_steps', 'cost', 'initial_pause_cost', 'running_pause_cost', 'resumption_cost', 'cancellation_cost'
        , 'line', 'process']
    n_steps: int
    """Number of steps needed to complete the manufacturing"""
    cost: float
    """Cost of manufacturing"""
    initial_pause_cost: float
    """Cost of pausing incurred only at the step a pause is started"""
    running_pause_cost: float
    """Running cost of pausing"""
    resumption_cost: float
    """Cost of resuming a process"""
    cancellation_cost: float
    """Cost of cancelling the process before the last step"""
    line: int
    """The line index"""
    process: int
    """The `Process` index"""

    @classmethod
    def from_manufacturing_profile(cls, profile: ManufacturingProfile, process2ind: Dict[Process, int]):
        return ManufacturingProfileCompiled(n_steps=profile.n_steps, cost=profile.cost
                                            , initial_pause_cost=profile.initial_pause_cost
                                            , running_pause_cost=profile.running_pause_cost
                                            , resumption_cost=profile.resumption_cost
                                            , cancellation_cost=profile.cancellation_cost
                                            , line=profile.line
                                            , process=process2ind[profile.process])


@dataclass
class ProductManufacturingInfo:
    """Gives full information about a manufacturing process that can generate or consume a product.

    See Also:
        `consuming` and `producing` of `Factory`
    """
    __slots__ = ['profile', 'quantity', 'step']
    profile: int
    """The `ManufacturingProfile` index"""
    quantity: int
    """The quantity generated/consumed by running this manufacturing info"""
    step: int
    """The step from the beginning at which the `Product` is received/consumed"""


@dataclass
class FinancialReport:
    """Reports that financial standing of an agent at a given time in the simulation"""
    agent: str
    """Agent ID"""
    step: int
    """Time of the report"""
    cash: float
    """Cash at hand"""
    liabilities: float
    """Total liabilities (loans)"""
    inventory: float
    """Value of everything in the inventory priced at catalog prices."""
    credit_rating: float
    """The agent's credit rating as a fraction of the maximum credit rating (1 indicates highest credit rating)."""

    @property
    def balance(self):
        """The balance of the agent defined as the difference between its available cash + inventory and its liabilities

        Remarks:

            - If the inventory was not calculated (due to having at least one product with unknown catalog price),
              it is used as zero in the equation.
        """
        return self.cash + self.inventory - self.liabilities if self.inventory is not None else self.cash - self.liabilities


@dataclass
class Loan:
    amount: float
    """Loan amount"""
    starts_at: int
    """The time-step at which payment starts"""
    total: float
    """The total to be paid including the amount + interests"""
    interest: float
    """The interest rate per step"""
    installment: float
    """The amount to be paid in one installment"""
    n_installments: int
    """The number of installments"""

    def __str__(self):
        return f'{self.amount} @ {self.interest} paid in {self.n_installments} [{self.installment} each] ' \
            f'for a total {self.total} [starts at {self.starts_at}]'

    class Java:
        implements = ['jnegmas.apps.scml.common.Loan']


RunningNegotiationInfo = namedtuple('RunningNegotiationInfo', ['negotiator', 'annotation', 'uuid', 'extra'])
"""Keeps track of running negotiations for an agent"""

NegotiationRequestInfo = namedtuple('NegotiationRequestInfo', ['partners', 'issues', 'annotation', 'uuid'
    , 'negotiator', 'extra'])
"""Keeps track to negotiation requests that an agent sent"""


@dataclass
class InsurancePolicy:
    premium: float
    contract: Contract
    at_time: int
    against: 'SCMLAgent'


@dataclass
class Factory:
    """Represents a factory within an SCML world. It is only accessed by the World so it need not be made public."""
    initial_storage: InitVar[Dict[int, int]]
    """Initial storage"""
    initial_wallet: InitVar[float] = 0.0
    """Initial Wallet"""
    id: str = field(default_factory=lambda: str(uuid.uuid4()), init=True)
    """Object name"""
    profiles: List[ManufacturingProfile] = field(default_factory=list)
    """A list of profiles used to initialize the factory"""
    max_storage: int = sys.maxsize
    """Maximum storage allowed in this factory"""
    min_storage: int = 0
    """Minimum allowed storage per product"""
    min_balance: int = 0
    """Minimum allowed balance"""
    initial_balance: float = field(init=False, default=0.0)
    """Initial balance of the factory"""
    _commands: np.array = field(init=False)
    """The production command currently running"""
    _line_schedules: np.array = field(init=False)
    _storage: Dict[int, int] = field(default_factory=lambda: defaultdict(int), init=False)
    """Mapping from product index to the amount available in the inventory"""
    _total_storage: int = field(init=False, default=0)
    """Total storage"""
    _wallet: float = field(default=0, init=False)
    """Money available for purchases"""
    _hidden_money: float = field(default=0, init=False)
    """Amount of money hidden by the agent"""
    _hidden_storage: Dict[int, int] = field(default_factory=lambda: defaultdict(int), init=False)
    """Mapping from product index to the amount hidden by the agent"""
    _loans: float = field(default=0.0, init=False)
    """The total money owned as loans"""
    _n_lines: int = field(init=False)
    """The number of lines in the factory, will be set using the `profiles` input"""
    _jobs: Dict[Tuple[int, int], Job] = field(default_factory=dict)
    """The jobs waiting to be run on the factory"""
    _next_step: int = field(init=False, default=0)
    """Current simulation step"""
    _carried_updates: FactoryStatusUpdate = field(init=False, default_factory=lambda: FactoryStatusUpdate.empty())
    """Carried updates from last executed command"""

    def __post_init__(self, initial_storage: Dict[int, int], initial_wallet=0.0):
        # no matter what are the line indices in the given profiles, the lines used by the factory
        # will be numbered from 0 to `n_lines` - 1
        if self.max_storage is None or self.max_storage < 0:
            self.max_storage = sys.maxsize
        given_lines = sorted(list(set(p.line for p in self.profiles)))
        mapping = dict(zip(given_lines, range(len(given_lines))))
        for profile in self.profiles:
            profile.line = mapping[profile.line]
        self._n_lines = len(given_lines)
        self._commands = np.array([RunningCommandInfo.do_nothing() for _ in range(self._n_lines)])
        self._line_schedules = np.ones(self._n_lines, dtype=int) * NO_PRODUCTION
        self._storage = defaultdict(int)
        self._total_storage = 0
        for k, v in initial_storage.items():
            self._storage[k] = v
            self._total_storage += v
        self._wallet = initial_wallet
        self._carried_updates = FactoryStatusUpdate.empty()
        self.initial_balance = initial_wallet

    @property
    def hidden_money(self) -> float:
        return self._hidden_money

    @property
    def hidden_storage(self) -> Dict[int, int]:
        return self._hidden_storage

    @property
    def n_lines(self) -> int:
        return self._n_lines

    @property
    def commands(self) -> np.array:
        return self._commands

    @property
    def line_schedules(self) -> np.array:
        return self._line_schedules

    @property
    def wallet(self) -> float:
        return self._wallet

    @property
    def storage(self) -> Dict[int, int]:
        return self._storage

    @property
    def loans(self) -> float:
        return self._loans

    @property
    def total_storage(self) -> int:
        # assert len(self._storage) == 0 or min(self._storage.values()) >= self.min_storage, f'min {min(self._storage.values()) if len(self._storage) > 0 else 0} < {self.min_storage} storage {self._storage}'
        # assert (len(self._storage) == 0 and self._total_storage == 0) or \
        #     self._total_storage == sum(self._storage.values()), f'Total storage {self._total_storage} but the sum is {sum(self._storage.values())}'
        return self._total_storage

    @property
    def balance(self) -> float:
        """The total balance of the factory"""
        return self._wallet - self._loans

    @property
    def next_step(self) -> int:
        return self._next_step

    def add_loan(self, total: float) -> None:
        self._loans += total

    def receive(self, payment: float) -> None:
        self.pay(-payment)

    def pay(self, payment: float) -> None:
        if self._wallet - payment < self.min_balance:
            raise ValueError(f'Cannot pay {payment} as  we have only {self._wallet}')
        self._wallet -= payment

    def transport_to(self, product: int, quantity: int) -> None:
        if self._storage[product] + quantity < self.min_storage:
            raise ValueError(f'Cannot transfer {quantity} of {product} as  we have only {self._storage[product]} '
                             f'(min {self.min_storage}, max {self.max_storage})')
        if self._total_storage + quantity > self.max_storage:
            raise ValueError(f'Cannot transfer {quantity} of {product} as  we have only {self._storage[product]} '
                             f'(min {self.min_storage}, max {self.max_storage})')
        self._storage[product] += quantity
        self._total_storage += quantity

    # @todo Schedulers and simulators do not know about transportation or transfer delays. They should
    # @todo Factory buy and sell functions do not take transportation and transfer delays into account

    def buy(self, product: int, quantity: int, price: float) -> None:
        if self._wallet < price or self._total_storage + quantity > self.max_storage:
            raise ValueError(f'Cannot buy {quantity} (total {self._total_storage}/{sum(self._storage.values())}) of '
                             f'{product} for {price} (wallet {self._wallet} / balance {self.balance})')
        self._wallet -= price
        self._storage[product] += quantity
        self._total_storage += quantity

    def sell(self, product: int, quantity: int, price: float) -> None:
        if self._storage[product] < quantity + self.min_storage:
            raise ValueError(f'Cannot sell {quantity} (have {self._storage[product]}) of '
                             f'{product} for {price} (wallet {self._wallet} / balance {self.balance})')
        self._storage[product] -= quantity
        self._total_storage -= quantity
        self._wallet += price

    def transport_from(self, product: int, quantity: int) -> None:
        self.transport_to(product=product, quantity=-quantity)

    def hide_funds(self, amount: float) -> None:
        to_hide = min(amount, self._wallet)
        self._hidden_money += to_hide
        self._wallet -= to_hide

    def hide_product(self, product: int, quantity: int) -> None:
        to_hide = min(quantity, self._storage.get(product, 0))
        self._hidden_storage[product] += to_hide
        self._storage[product] -= to_hide

    def unhide_funds(self, amount: float) -> None:
        to_hide = min(amount, self._hidden_money)
        self._hidden_money -= to_hide
        self._wallet += to_hide

    def unhide_product(self, product: int, quantity: int) -> None:
        to_hide = min(quantity, self._hidden_storage.get(product, 0))
        self._hidden_storage[product] -= to_hide
        self._storage[product] += to_hide

    def schedule(self, job: Job, override=False) -> None:
        """
        Schedules the given job at its `time` and `line` optionally overriding whatever was already scheduled
        Args:
            job:
            override:

        Returns:
            Success/failure
        """
        # you can only schedule jobs at the following simulation step
        t, line, profile = job.time, job.line, self.profiles[job.profile]
        if job.action in ('run', 'start'):
            line = profile.line
        if t < self._next_step - 1 or line >= self._n_lines or line < 0:
            raise ValueError(f'cannot schedule at time {t} (current {self._next_step - 1}) on line {line} '
                             f'of {self._n_lines}')
        existing_job = self._jobs.get((t, line), None)
        if existing_job is None:
            self._jobs[(t, line)] = job
            return
        if existing_job.is_cancelling(job):
            del self._jobs[(t, line)]
            return
        if not override:
            raise ValueError(f'Job {str(existing_job)} is scheduled at {t} and overriding is not allowed')
        self._jobs[(t, line)] = job

    def _apply_updates(self, updates: FactoryStatusUpdate) -> None:
        if updates.balance != 0.0:
            self._wallet += updates.balance
        if updates.storage is not None:
            for k, v in updates.storage.items():
                self._storage[k] += v
                self._total_storage += v

    def step(self) -> List[ProductionReport]:
        reports = []
        if not self._carried_updates.is_empty:
            reports.append(ProductionReport(line=-1, started=None, continuing=None, finished=None, failure=None
                                            , updates=self._carried_updates))
            self._apply_updates(self._carried_updates)
            self._carried_updates = FactoryStatusUpdate.empty()
        for line in range(self._n_lines):
            # step the current production process
            if self._commands[line].ended_before(self._next_step):
                self._commands[line].action = 'none'
            report = self._step_line(line=line)
            reports.append(report)
            self._apply_updates(report.updates)
        self._next_step += 1
        return reports

    def _run(self, profile: ManufacturingProfile, override=True) -> None:
        """running is executed at the beginning of the step t

        Args:
            profile: the profile to start giving both the line and process
            override: If true, override any running processes paying cancellation cost for these processes

        Remarks:

            - The output of a process that runs from step t to step t + n - 1 will only be in storage at step t + n

        """

        # if I am not allowed to override, then this command has no effect and I return an empty status update
        t = self._next_step
        line = profile.line
        running_command = self._commands[line]
        if not running_command.is_none and not override:
            return
        process = profile.process
        n, cost = profile.n_steps, profile.cost
        updates = defaultdict(lambda: FactoryStatusUpdate.empty())
        command = RunningCommandInfo(action='run', profile=profile, beg=t, end=t + n, updates=updates, paused=False
                                     , step=0)
        for need in process.inputs:
            updates[int(math.floor(need.step * n))].storage[need.product] -= need.quantity
        for output in process.outputs:
            updates[int(math.ceil(output.step * n))].storage[output.product] += output.quantity
        updates[0].balance -= cost

        # cancel the running command by stopping it and then run the new command
        if not running_command.is_none:
            self._stop(line=profile.line)
        self._commands[line] = command
        self._line_schedules[line] = process.id

    def _pause(self, line: int) -> None:
        """pausing is executed at the end of the step

        Args:
            line: the line on which the process is running

        Returns:

            Optional[Dict[int, FactoryStatusUpdate]]: The status updated for all times that need to be updated to cancel
            the command if it is not None. If None is returned then scheduling failed.

        Remarks:

            - Not implemented yet
            - pausing when nothing is running is not an error and will return an empty status update

        """
        running_command = self._commands[line]
        if running_command.is_none:
            return
        running_command.updates[running_command.step].combine(
            FactoryStatusUpdate(balance=-running_command.profile.initial_pause_cost, storage={}))
        running_command.paused = True

    def _resume(self, line: int) -> None:
        """resumption is executed at the end of the step (starting next step count down)


        Args:
            line: the line on which the process is running

        Returns:

            Optional[Dict[int, FactoryStatusUpdate]]: The status updated for all times that need to be updated to cancel
            the command if it is not None. If None is returned then scheduling failed.

        Remarks:

            - Not implemented yet
            - resuming when nothing is paused is not an error and will return an empty status update

        """
        running_command = self._commands[line]
        if running_command.is_none:
            return
        profile = running_command.profile
        running_command.updates[running_command.step].combine(FactoryStatusUpdate(balance=-profile.resumption_cost
                                                                                  , storage={}))
        running_command.paused = False

    def _stop(self, line: int) -> None:
        """stopping is executed at the beginning of the current step

        Args:
            line: the line on which the process is running

        Returns:

            Optional[Dict[int, FactoryStatusUpdate]]: The status updated for all times that need to be updated to cancel
            the command if it is not None. If None is returned then scheduling failed.

        Remarks:

            - stopping when nothing is running is not an error and will just return an empty schedule
        """

        # stopping a no-action command is always successful
        running_command = self._commands[line]
        if running_command.is_none:
            return
        profile = running_command.profile
        t = self._next_step
        running_command.paused = False
        running_command.end = t + 1
        running_command.updates = {running_command.step: running_command.updates[running_command.step].combine
        (FactoryStatusUpdate(balance=-profile.cancellation_cost, storage={}))}

    def _step_line(self, line: int) -> ProductionReport:
        """
        Steps the line to the time-step `t` assuming that it is already stepped to time-step t-1 given the storage

        Args:
            line: the line to step

        Returns:
            ProductionReport
        """
        t = self._next_step
        running_command = self._commands[line]
        job = self._jobs.get((t, line), None)
        updates = FactoryStatusUpdate.empty()
        if job is None and running_command.is_none:
            return ProductionReport(updates=updates, continuing=None, started=None, finished=None
                                    , failure=None, line=line)
        if job is not None:
            if job.action in ('run', 'start'):
                self._run(profile=self.profiles[job.profile], override=job.override)
            elif job.action == 'pause':
                self._pause(line=job.line)
            elif job.action == 'resume':
                self._resume(line=job.line)
            elif job.action == 'stop':
                self._stop(line=job.line)
            else:
                raise ValueError(f'action {job.action} is unknown')
            del self._jobs[(t, line)]

        # now all updates in the command are correct except for the running pause cost which we add here
        running_command = self._commands[line]
        profile = running_command.profile
        if running_command.paused:
            running_command.end += 1
            running_command.updates = {k + 1: v for k, v in running_command.updates.items()}
            running_command.updates[running_command.step].combine(
                FactoryStatusUpdate(balance=-profile.running_pause_cost, storage={}))
        updates = running_command.updates.get(running_command.step, None)
        if not running_command.paused:
            running_command.step += 1
        if updates is None or updates.is_empty:
            return ProductionReport(updates=FactoryStatusUpdate.empty()
                                    , continuing=running_command if running_command.beg < t else None
                                    , started=running_command if running_command.beg == t else None
                                    , finished=running_command if running_command.end <= t + 1 else None
                                    , failure=None, line=line)
        if updates is not None:
            del running_command.updates[running_command.step - 1]
        available_storage = self.max_storage - self._total_storage
        missing_inputs = []
        missing_money = 0
        failed = False
        missing_space = 0
        if updates.balance < 0 and self._wallet < -updates.balance:
            failed = True
            missing_money = -updates.balance - self._wallet
        for product_id, quantity in updates.storage.items():
            if quantity < 0 and self._storage.get(product_id, 0) < -quantity:
                failed = True
                missing_inputs.append(MissingInput(product=product_id, quantity=-quantity))
            elif quantity > 0:
                available_storage -= quantity
                if available_storage < 0:
                    failed = True
                    missing_space -= available_storage
                    available_storage = 0
        if failed:
            running_command.action = 'none'
            failure = ProductionFailure(line=line, command=running_command, missing_money=missing_money
                                        , missing_inputs=missing_inputs, missing_space=missing_space)
            return ProductionReport(updates=FactoryStatusUpdate.empty()
                                    , continuing=running_command if running_command.beg < t else None
                                    , started=running_command if running_command.beg == t else None
                                    , finished=running_command if running_command.end <= t + 1 else None
                                    , failure=failure, line=line)
        if running_command.ended_before(t + 1):
            self._carried_updates.combine(running_command.updates.get(running_command.step
                                                                      , FactoryStatusUpdate.empty()))
        return ProductionReport(updates=updates
                                , continuing=running_command if running_command.beg < t else None
                                , started=running_command if running_command.beg == t else None
                                , finished=running_command if running_command.end <= t + 1 else None
                                , failure=None, line=line)



import pytest

from negmas import Negotiator


def test_basic_agent_has_good_name():
    class MyAgent(Negotiator):
        def ufun(self, negotiation_id):
            return None

        def isin(self, negotiation_id) -> bool:
            return False

        def leave(self, negotiation) -> bool:
            return False

        def enter(self, negotiation, ufun) -> bool:
            return False

    x = MyAgent()

    assert x.isin(None) is False
    assert x.leave(None) is False
    assert x.enter(None, None) is False


if __name__ == '__main__':
    pytest.main(args=[__file__])



import itertools
from abc import abstractmethod, ABC
from collections import defaultdict

from negmas import Contract, Breach, Negotiator, MechanismState, AgentMechanismInterface, RenegotiationRequest, \
    JavaSAONegotiator, UtilityFunction, JavaUtilityFunction
from negmas.apps.scml.simulators import FactorySimulator, FastFactorySimulator
from negmas.apps.scml.simulators import storage_as_array, temporary_transaction
from negmas.common import NamedObject
from negmas.events import Notification
from negmas.helpers import get_class
from negmas.helpers import snake_case
from negmas.java import JavaCallerMixin, to_java, from_java, to_dict_for_java, java_link, PYTHON_CLASS_IDENTIFIER
from negmas.outcomes import Issue, Outcome
from negmas.sao import AspirationNegotiator
from negmas.utilities import UtilityValue, normalize
from .agent import SCMLAgent
from .awi import _ShadowSCMLAWI, SCMLAWI
from .common import SCMLAgreement, Factory, INVALID_UTILITY, CFP, Loan, ProductionFailure, FinancialReport
from .consumers import ScheduleDrivenConsumer, ConsumptionProfile
from .schedulers import Scheduler, ScheduleInfo, GreedyScheduler

if True:
    from typing import Dict, Iterable, Any, Callable, Collection, Type, List, Optional, Union

__all__ = [
    'FactoryManager', 'DoNothingFactoryManager', 'GreedyFactoryManager',
    'JavaFactoryManager', 'JavaDoNothingFactoryManager', 'JavaGreedyFactoryManager', 'JavaDummyMiddleMan',
]


class FactoryManager(SCMLAgent, ABC):
    """Base factory manager class that will be inherited by participant negmas in ANAC 2019"""

    @property
    def type_name(self):
        return super().type_name.replace('_factory_manager', '')

    def __init__(self, name=None, simulator_type: Union[str, Type[FactorySimulator]] = FastFactorySimulator):
        super().__init__(name=name)
        self.transportation_delay = 0
        self.simulator: Optional[FactorySimulator] = None
        self.simulator_type: Type[FactorySimulator] = get_class(simulator_type, scope=globals())
        self.current_step = 0
        self.max_storage: int = 0

    def init_(self):
        state: Factory = self.awi.state
        self.current_step = state.next_step
        self.max_storage = state.max_storage
        self.simulator = self.simulator_type(initial_wallet=state.wallet, initial_storage=state.storage
                                             , n_steps=self.awi.n_steps, n_products=len(self.awi.products)
                                             , profiles=state.profiles, max_storage=self.max_storage)
        super().init_()

    def step_(self):
        state = self.awi.state
        self.simulator.set_state(self.current_step, wallet=state.wallet, loans=state.loans
                                 , storage=storage_as_array(state.storage, n_products=len(self.products))
                                 , line_schedules=state.line_schedules)
        self.current_step += 1
        self.step()

    @abstractmethod
    def on_production_failure(self, failures: List[ProductionFailure]) -> None:
        """
        Called with a list of `ProductionFailure` records on production failure

        Args:
            failures:

        Returns:

        """

    class Java:
        implements = ['jnegmas.apps.scml.factory_managers.FactoryManager']


class DoNothingFactoryManager(FactoryManager):
    """The default factory manager that will be implemented by the committee of ANAC-SCML 2019"""

    def on_new_report(self, report: FinancialReport):
        pass

    def init(self):
        pass

    def step(self):
        pass

    def on_neg_request_rejected(self, req_id: str, by: Optional[List[str]]):
        pass

    def on_neg_request_accepted(self, req_id: str, mechanism: AgentMechanismInterface):
        pass

    def on_negotiation_failure(self, partners: List[str], annotation: Dict[str, Any],
                               mechanism: AgentMechanismInterface, state: MechanismState) -> None:
        pass

    def on_negotiation_success(self, contract: Contract, mechanism: AgentMechanismInterface) -> None:
        pass

    def on_contract_signed(self, contract: Contract) -> None:
        pass

    def on_contract_cancelled(self, contract: Contract, rejectors: List[str]) -> None:
        pass

    def sign_contract(self, contract: Contract) -> Optional[str]:
        return self.id

    def on_contract_nullified(self, contract: Contract, bankrupt_partner: str, compensation: float) -> None:
        pass

    def on_agent_bankrupt(self, agent_id: str) -> None:
        pass

    def confirm_partial_execution(self, contract: Contract, breaches: List[Breach]) -> bool:
        return True

    def on_remove_cfp(self, cfp: 'CFP') -> None:
        pass

    def on_production_failure(self, failures: List[ProductionFailure]) -> None:
        pass

    def respond_to_negotiation_request(self, cfp: "CFP", partner: str) -> Optional[Negotiator]:
        return None

    def confirm_contract_execution(self, contract: Contract) -> bool:
        return True

    def set_renegotiation_agenda(self, contract: Contract, breaches: List[Breach]) -> Optional[RenegotiationRequest]:
        return None

    def respond_to_renegotiation_request(self, contract: Contract, breaches: List[Breach]
                                         , agenda: RenegotiationRequest) -> Optional[Negotiator]:
        return None

    def confirm_loan(self, loan: Loan, bankrupt_if_rejected: bool) -> bool:
        """called by the world manager to confirm a loan if needed by the buyer of a contract that is about to be
        breached"""
        return bankrupt_if_rejected

    def on_new_cfp(self, cfp: 'CFP') -> None:
        pass


class GreedyFactoryManager(DoNothingFactoryManager):
    """The default factory manager that will be implemented by the committee of ANAC-SCML 2019"""

    def on_production_failure(self, failures: List[ProductionFailure]) -> None:
        pass

    def confirm_loan(self, loan: Loan, bankrupt_if_rejected: bool) -> bool:
        return bankrupt_if_rejected

    def confirm_contract_execution(self, contract: Contract) -> bool:
        return True

    def set_renegotiation_agenda(self, contract: Contract, breaches: List[Breach]) -> Optional[RenegotiationRequest]:
        return None

    def respond_to_renegotiation_request(self, contract: Contract, breaches: List[Breach],
                                         agenda: RenegotiationRequest) -> Optional[Negotiator]:
        return None

    def __init__(self, name=None, simulator_type: Union[str, Type[FactorySimulator]] = FastFactorySimulator
                 , scheduler_type: Union[str, Type[Scheduler]] = GreedyScheduler
                 , scheduler_params: Optional[Dict[str, Any]] = None
                 , optimism: float = 0.0
                 , negotiator_type: Union[str, Type[Negotiator]] = 'negmas.sao.AspirationNegotiator'
                 , negotiator_params: Optional[Dict[str, Any]] = None
                 , n_retrials=5, use_consumer=True, reactive=True, sign_only_guaranteed_contracts=False
                 , riskiness=0.0, max_insurance_premium: float = -1.0):
        super().__init__(name=name, simulator_type=simulator_type)
        self.negotiator_type = get_class(negotiator_type, scope=globals())
        self.negotiator_params = negotiator_params if negotiator_params is not None else {}
        self.optimism = optimism
        self.ufun_factory: Union[Type[NegotiatorUtility], Callable[[Any, Any], NegotiatorUtility]]
        if optimism < 1e-6:
            self.ufun_factory = PessimisticNegotiatorUtility
        elif optimism > 1 - 1e-6:
            self.ufun_factory = OptimisticNegotiatorUtility
        else:
            self.ufun_factory: NegotiatorUtility = lambda agent, annotation: \
                AveragingNegotiatorUtility(agent=agent, annotation=annotation, optimism=self.optimism)
        self.max_insurance_premium = max_insurance_premium
        self.n_retrials = n_retrials
        self.n_neg_trials: Dict[str, int] = defaultdict(int)
        self.consumer = None
        self.use_consumer = use_consumer
        self.reactive = reactive
        self.sign_only_guaranteed_contracts = sign_only_guaranteed_contracts
        self.contract_schedules: Dict[str, ScheduleInfo] = {}
        self.riskiness = riskiness
        self.negotiation_margin = int(round(n_retrials * max(0.0, 1.0 - riskiness)))
        self.scheduler_type: Type[Scheduler] = get_class(scheduler_type, scope=globals())
        self.scheduler: Scheduler = None
        self.scheduler_params: Dict[str, Any] = scheduler_params if scheduler_params is not None else {}

    def total_utility(self, contracts: Collection[Contract] = ()) -> float:
        """Calculates the total utility for the agent of a collection of contracts"""
        if self.scheduler is None:
            raise ValueError('Cannot calculate total utility without a scheduler')
        min_concluded_at = self.awi.current_step
        min_sign_at = min_concluded_at + self.awi.default_signing_delay
        with temporary_transaction(self.scheduler):
            schedule = self.scheduler.schedule(contracts=contracts, assume_no_further_negotiations=False
                                               , ensure_storage_for=self.transportation_delay
                                               , start_at=min_sign_at)
        if not schedule.valid:
            return INVALID_UTILITY
        return schedule.final_balance

    def init(self):
        if self.use_consumer:
            # @todo add the parameters of the consumption profile as parameters of the greedy factory manager
            self.consumer: ScheduleDrivenConsumer = ScheduleDrivenConsumer(profiles=dict(zip(self.consuming.keys()
                                                                                             , (ConsumptionProfile(
                    schedule=[_] * self.awi.n_steps)
                                                                                                 for _ in
                                                                                                 itertools.repeat(0))))
                                                                           , consumption_horizon=self.awi.n_steps,
                                                                           immediate_cfp_update=True
                                                                           , name=self.name)
            self.consumer.id = self.id
            self.consumer.awi = self.awi
            self.consumer.init_()
        self.scheduler = self.scheduler_type(manager_id=self.id, awi=self.awi
                                             , max_insurance_premium=self.max_insurance_premium
                                             , **self.scheduler_params)
        self.scheduler.init(simulator=self.simulator, products=self.products, processes=self.processes
                            , producing=self.producing, profiles=self.compiled_profiles)

    def respond_to_negotiation_request(self, cfp: "CFP", partner: str) -> Optional[Negotiator]:
        if self.awi.is_bankrupt(partner):
            return None
        if self.use_consumer:
            return self.consumer.respond_to_negotiation_request(cfp=cfp, partner=partner)
        else:
            neg = self.negotiator_type(name=self.name + '*' + partner, **self.negotiator_params)
            neg.utility_function = normalize(self.ufun_factory(self, self._create_annotation(cfp=cfp)),
                                             outcomes=cfp.outcomes, infeasible_cutoff=0)
            return neg

    def on_negotiation_success(self, contract: Contract, mechanism: AgentMechanismInterface):
        if self.use_consumer:
            self.consumer.on_negotiation_success(contract, mechanism)

    def on_negotiation_failure(self, partners: List[str], annotation: Dict[str, Any], mechanism: AgentMechanismInterface
                               , state: MechanismState) -> None:
        if self.use_consumer:
            self.consumer.on_negotiation_failure(partners, annotation, mechanism, state)
        cfp = annotation['cfp']
        thiscfp = self.awi.bb_query(section='cfps', query=cfp.id, query_keys=True)
        if cfp.publisher != self.id and thiscfp is not None and len(thiscfp) > 0 \
            and self.n_neg_trials[cfp.id] < self.n_retrials:
            self.awi.logdebug(f'Renegotiating {self.n_neg_trials[cfp.id]} on {cfp}')
            self.on_new_cfp(cfp=annotation['cfp'])

    def _execute_schedule(self, schedule: ScheduleInfo, contract: Contract) -> None:
        if self.simulator is None:
            raise ValueError('No factory simulator is defined')
        awi: SCMLAWI = self.awi
        total = contract.agreement['unit_price'] * contract.agreement['quantity']
        product = contract.annotation['cfp'].product
        if contract.annotation['buyer'] == self.id:
            self.simulator.buy(product=product, quantity=contract.agreement['quantity']
                               , price=total, t=contract.agreement['time']
                               )
            if total <= 0 or self.max_insurance_premium < 0.0 or contract is None:
                return
            premium = awi.evaluate_insurance(contract=contract)
            if premium is None:
                return
            relative_premium = premium / total
            if relative_premium <= self.max_insurance_premium:
                awi.buy_insurance(contract=contract)
                self.simulator.pay(premium, self.awi.current_step)
            return
        # I am a seller
        self.simulator.sell(product=product, quantity=contract.agreement['quantity']
                            , price=total, t=contract.agreement['time'])
        for job in schedule.jobs:
            if job.action == 'run':
                awi.schedule_job(job, contract=contract)
            elif job.action == 'stop':
                awi.stop_production(line=job.line, step=job.time, contract=contract, override=job.override)
            else:
                awi.schedule_job(job, contract=contract)
            self.simulator.schedule(job=job, override=False)
        for need in schedule.needs:
            if need.quantity_to_buy <= 0:
                continue
            product_id = need.product
            # self.simulator.reserve(product=product_id, quantity=need.quantity_to_buy, t=need.step)
            if self.use_consumer:
                self.consumer.profiles[product_id].schedule[need.step] += need.quantity_to_buy
                self.consumer.register_product_cfps(p=product_id, t=need.step
                                                    , profile=self.consumer.profiles[product_id])
                continue
            product = self.products[product_id]
            if product.catalog_price is None:
                price_range = (0.0, 100.0)
            else:
                price_range = (0.5, 1.5 * product.catalog_price)
            # @todo check this. This error is raised sometimes
            if need.step < awi.current_step:
                continue
                # raise ValueError(f'need {need} at {need.step} while running at step {awi.current_step}')
            time = need.step if self.max_storage is not None else (awi.current_step, need.step)
            cfp = CFP(is_buy=True, publisher=self.id, product=product_id
                      , time=time, unit_price=price_range
                      , quantity=(1, int(1.1 * need.quantity_to_buy)))
            awi.register_cfp(cfp)

    def sign_contract(self, contract: Contract):
        if any(self.awi.is_bankrupt(partner) for partner in contract.partners):
            return None
        signature = self.id
        with temporary_transaction(self.scheduler):
            schedule = self.scheduler.schedule(assume_no_further_negotiations=False, contracts=[contract]
                                               , ensure_storage_for=self.transportation_delay
                                               , start_at=self.awi.current_step + 1)

        if self.sign_only_guaranteed_contracts and (not schedule.valid or len(schedule.needs) > 1):
            self.awi.logdebug(f'{self.name} refused to sign contract {contract.id} because it cannot be scheduled')
            return None
        # if schedule.final_balance <= self.simulator.final_balance:
        #     self.awi.logdebug(f'{self.name} refused to sign contract {contract.id} because it is not expected '
        #                       f'to lead to profit')
        #     return None
        if schedule.valid:
            profit = schedule.final_balance - self.simulator.final_balance
            self.awi.logdebug(f'{self.name} singing contract {contract.id} expecting '
                              f'{-profit if profit < 0 else profit} {"loss" if profit < 0 else "profit"}')
        else:
            self.awi.logdebug(f'{self.name} singing contract {contract.id} expecting breach')
            return None

        self.contract_schedules[contract.id] = schedule
        return signature

    def on_contract_signed(self, contract: Contract):
        if contract.annotation['buyer'] == self.id and self.use_consumer:
            self.consumer.on_contract_signed(contract)
        schedule = self.contract_schedules[contract.id]
        if schedule is not None and schedule.valid:
            self._execute_schedule(schedule=schedule, contract=contract)
        if contract.annotation['buyer'] != self.id or not self.use_consumer:
            for negotiation in self._running_negotiations.values():
                self.notify(negotiation.negotiator, Notification(type='ufun_modified', data=None))

    def _process_buy_cfp(self, cfp: 'CFP') -> None:
        if self.awi.is_bankrupt(cfp.publisher):
            return None
        if self.simulator is None or not self.can_expect_agreement(cfp=cfp, margin=self.negotiation_margin):
            return
        if not self.can_produce(cfp=cfp):
            return
        if self.negotiator_type == AspirationNegotiator:
            neg = self.negotiator_type(assume_normalized=True, name=self.name + '>' + cfp.publisher)
        else:
            neg = self.negotiator_type(name=self.name + '>' + cfp.publisher)
        self.request_negotiation(negotiator=neg, cfp=cfp
                                 , ufun=normalize(self.ufun_factory(self, self._create_annotation(cfp=cfp))
                                                  , outcomes=cfp.outcomes, infeasible_cutoff=-1500))

    def _process_sell_cfp(self, cfp: 'CFP'):
        if self.awi.is_bankrupt(cfp.publisher):
            return None
        if self.use_consumer:
            self.consumer.on_new_cfp(cfp=cfp)

    def on_new_cfp(self, cfp: 'CFP') -> None:
        if not self.reactive:
            return
        if cfp.satisfies(query={'is_buy': True, 'products': list(self.producing.keys())}):
            self._process_buy_cfp(cfp)
        if cfp.satisfies(query={'is_buy': False, 'products': list(self.consuming.keys())}):
            self._process_sell_cfp(cfp)

    def step(self):
        if self.use_consumer:
            self.consumer.step()
        if self.reactive:
            return
        # 0. remove all my CFPs
        # self.awi.bb_remove(section='cfps', query={'publisher': self})

        # respond to interesting CFPs
        # todo: should check time and sort products by interest etc
        cfps = self.awi.bb_query(section='cfps'
                                 , query={'products': self.producing.keys(), 'is_buy': True})
        if cfps is None:
            return
        for cfp in cfps.values():
            self._process_buy_cfp(cfp)

    def can_produce(self, cfp: CFP, assume_no_further_negotiations=False) -> bool:
        """Whether or not we can produce the required item in time"""
        if cfp.product not in self.producing.keys():
            return False
        agreement = SCMLAgreement(time=cfp.max_time, unit_price=cfp.max_unit_price, quantity=cfp.min_quantity)
        min_concluded_at = self.awi.current_step + 1 - int(self.immediate_negotiations)
        min_sign_at = min_concluded_at + self.awi.default_signing_delay
        if cfp.max_time < min_sign_at + 1:  # 1 is minimum time to produce the product
            return False
        with temporary_transaction(self.scheduler):
            schedule = self.scheduler.schedule(contracts=[Contract(partners=[self.id, cfp.publisher]
                                                                   , agreement=agreement
                                                                   , annotation=self._create_annotation(cfp=cfp)
                                                                   , issues=cfp.issues, signed_at=min_sign_at
                                                                   , concluded_at=min_concluded_at)]
                                               , ensure_storage_for=self.transportation_delay
                                               , assume_no_further_negotiations=assume_no_further_negotiations
                                               , start_at=min_sign_at)
        return schedule.valid and self.can_secure_needs(schedule=schedule, step=self.awi.current_step)

    def can_secure_needs(self, schedule: ScheduleInfo, step: int):
        """
        Finds if it is possible in principle to arrange these needs at the given time.

        Args:
            schedule:
            step:

        Returns:

        """
        needs = schedule.needs
        if len(needs) < 1:
            return True
        for need in needs:
            if need.quantity_to_buy > 0 and need.step < step + 1 - int(self.immediate_negotiations):  # @todo check this
                return False
        return True


TotalUtilityFun = Callable[[Collection[Contract]], float]


class NegotiatorUtility(UtilityFunction):
    """The utility function of a negotiator."""

    def __init__(self, agent: GreedyFactoryManager, annotation: Dict[str, Any], name: Optional[str] = None):
        if name is None:
            name = agent.name + '*' + '*'.join(_ for _ in annotation['partners'] if _ != agent.id)
        super().__init__(name=name)
        self.agent = agent
        self.annotation = annotation

    def _contracts(self, agreements: Iterable[SCMLAgreement]) -> Collection[Contract]:
        """Converts agreements/outcomes into contracts"""
        if self.ami is None:
            raise ValueError('No annotation is stored (No mechanism info)')
        annotation = self.ami.annotation
        return [Contract(partners=annotation['partners'], agreement=a, annotation=annotation, issues=self.ami.issues)
                for a in agreements]

    def _contract(self, agreement: SCMLAgreement) -> Contract:
        """Converts an agreement/outcome into a contract"""
        annotation = self.annotation
        return Contract(partners=annotation['partners'], agreement=agreement
                        , annotation=annotation, issues=annotation['cfp'].issues)

    def _free_sale(self, agreement: SCMLAgreement) -> bool:
        return self.annotation['seller'] == self.agent.id and agreement['unit_price'] < 1e-6

    def __call__(self, outcome: Outcome) -> Optional[UtilityValue]:
        if isinstance(outcome, dict):
            return self.call(agreement=SCMLAgreement(**outcome))
        if isinstance(outcome, SCMLAgreement):
            return self.call(agreement=outcome)
        raise ValueError(f'Outcome: {outcome} cannot be converted to an SCMLAgreement')

    @abstractmethod
    def call(self, agreement: SCMLAgreement) -> Optional[UtilityValue]:
        """Called to evaluate a agreement"""

    def xml(self, issues: List[Issue]) -> str:
        return 'NegotiatorUtility has not xml representation'


class PessimisticNegotiatorUtility(NegotiatorUtility):
    """The utility function of a negotiator that assumes other negotiations currently open will fail."""

    def call(self, agreement: SCMLAgreement) -> Optional[UtilityValue]:
        """An offer will be a tuple of one value which in turn will be a list of contracts"""
        if self._free_sale(agreement):
            return INVALID_UTILITY
        # contracts = self.agent.contracts
        # hypothetical = list(contracts)
        # hypothetical.append(self._contract(agreement))
        hypothetical = [self._contract(agreement)]
        base_util = self.agent.simulator.final_balance
        hypothetical = self.agent.total_utility(hypothetical)
        if hypothetical < 0:
            return INVALID_UTILITY
        return hypothetical - base_util


class OptimisticNegotiatorUtility(NegotiatorUtility):
    """The utility function of a negotiator that assumes other negotiations currently open will succeed."""

    def call(self, agreement: SCMLAgreement) -> Optional[UtilityValue]:
        if self._free_sale(agreement):
            return INVALID_UTILITY
        # contracts = self.agent.contracts
        # hypothetical = list(contracts)
        # hypothetical.append(self._contract(agreement))
        hypothetical = [self._contract(agreement)]
        for negotiation in self.agent.running_negotiations:  # type: ignore
            negotiator = negotiation.negotiator
            current_offer = negotiator.my_last_proposal
            if current_offer is not None:
                hypothetical.append(self._contract(current_offer))
        base_util = self.agent.simulator.final_balance
        hypothetical = self.agent.total_utility(list(hypothetical))
        if hypothetical < 0:
            return INVALID_UTILITY
        return hypothetical - base_util


class AveragingNegotiatorUtility(NegotiatorUtility):
    """A utility function that combines optimistic and pessimistic evaluators linearly using adjustable weight"""

    def __init__(self, agent: GreedyFactoryManager, annotation: Dict[str, Any], name: Optional[str] = None
                 , optimism: float = 0.5):
        NamedObject.__init__(self=self, name=name)
        self.optimism = optimism
        self.optimistic = OptimisticNegotiatorUtility(agent=agent, annotation=annotation)
        self.pessimistic = PessimisticNegotiatorUtility(agent=agent, annotation=annotation)

    def call(self, agreement: SCMLAgreement) -> Optional[UtilityValue]:
        if self._free_sale(agreement):
            return INVALID_UTILITY
        opt, pess = self.optimistic(agreement), self.pessimistic(agreement)
        if opt is None or pess is None:
            return None
        return self.optimism * opt + (1 - self.optimism) * pess


class JavaFactoryManager(FactoryManager, JavaCallerMixin):
    """Allows factory managers implemented in Java (using jnegmas) to participate in SCML worlds.

    Objects of this class is used to represent a java object to the python environment. This means that they *MUST* have
    the same interface as a python class (first class in the inheritance list). The `JavaCallerMixin` is used to enable
    it to connect to the java object it is representing.

    """

    @property
    def type_name(self):
        """Overrides type name to give the internal java type name"""
        return 'j' + snake_case(self._java_class_name.replace('jnegmas.apps.scml.factory_managers.', '').replace(
            'FactoryManager', ''))

    @property
    def awi(self):
        return self._awi

    @awi.setter
    def awi(self, value):
        self._awi = value
        if self.python_shadow is not self:
            self.python_shadow._awi = value
        self.java_awi = _ShadowSCMLAWI(value)
        self._java_object.setAWI(self.java_awi)

    def init(self):
        if self.python_shadow is not self:
            self.python_shadow.simulator = self.simulator
        self._java_object.setSimulator(self.simulator)
        self._java_object.init()

    def step(self):
        return self._java_object.init()

    def on_neg_request_rejected(self, req_id: str, by: Optional[List[str]]):
        return self._java_object.onNegRequestRejected(req_id, by)

    def on_neg_request_accepted(self, req_id: str, mechanism: AgentMechanismInterface):
        return self._java_object.onNegRequestAccepted(req_id, java_link(mechanism, map=to_dict_for_java(mechanism
                                                                                                        , add_type_field=False)
                                                                        , copyable=False))

    def on_new_cfp(self, cfp: 'CFP'):
        return from_java(self._java_object.onNewCFP(to_java(cfp)))

    def on_remove_cfp(self, cfp: 'CFP'):
        return self._java_object.onRemoveCFP(to_java(cfp))

    def on_contract_nullified(self, contract: Contract, bankrupt_partner: str, compensation: float) -> None:
        self._java_object.onContractNullified(to_java(contract), bankrupt_partner, compensation)

    def on_agent_bankrupt(self, agent_id: str) -> None:
        self._java_object.onAgentBankrupt(agent_id)

    def confirm_partial_execution(self, contract: Contract, breaches: List[Breach]) -> bool:
        return self._java_object.confirmParialExecution(to_java(contract), to_java(breaches))

    def on_production_failure(self, failures: List[ProductionFailure]) -> None:
        return self._java_object.onProductionFailure(to_java(failures))

    def confirm_loan(self, loan: Loan, bankrupt_if_rejected: bool) -> bool:
        return self._java_object.confirmLoan(to_java(loan), bankrupt_if_rejected)

    def confirm_contract_execution(self, contract: Contract) -> bool:
        return self._java_object.confirmContractExecution(to_java(contract))

    def respond_to_negotiation_request(self, cfp: "CFP", partner: str) -> Optional[Negotiator]:
        result = self._java_object.respondToNegotiationRequest(to_java(cfp), partner)
        if result is None:
            return result
        return JavaSAONegotiator(java_object=result, java_class_name=None)

    def on_negotiation_failure(self, partners: List[str], annotation: Dict[str, Any], mechanism: AgentMechanismInterface
                               , state: MechanismState) -> None:
        return self._java_object.onNegotiationFailure(to_java(partners), annotation, java_link(mechanism)
                                                      , to_java(state))

    def on_negotiation_success(self, contract: Contract, mechanism: AgentMechanismInterface) -> None:
        return self._java_object.onNegotiationSuccess(to_java(contract), java_link(mechanism))

    def on_contract_signed(self, contract: Contract) -> None:
        return self._java_object.onContractSigned(to_java(contract))

    def on_contract_cancelled(self, contract: Contract, rejectors: List[str]) -> None:
        return self._java_object.onContractCancelled(to_java(contract), to_java(rejectors))

    def on_new_report(self, report: FinancialReport):
        pass

    def sign_contract(self, contract: Contract) -> Optional[str]:
        return from_java(self._java_object.signContract(to_java(contract)))

    def set_renegotiation_agenda(self, contract: Contract, breaches: List[Breach]) -> Optional[RenegotiationRequest]:
        return from_java(self._java_object.setRenegotiationAgenda(to_java(contract), to_java(breaches)))

    def respond_to_renegotiation_request(self, contract: Contract, breaches: List[Breach],
                                         agenda: RenegotiationRequest) -> Optional[Negotiator]:
        return from_java(self._java_object.respondToRenegotiationRequest(to_java(contract), to_java(breaches)
                                                                         , to_java(agenda)))

    # handy constructors
    @classmethod
    def do_nothing_manager(cls):
        return JavaFactoryManager(java_class_name='jnegmas.apps.scml.factory_managers.DoNothingFactoryManager')

    @classmethod
    def greedy_manager(cls):
        return JavaFactoryManager(java_class_name='jnegmas.apps.scml.factory_managers.GreedyFactoryManager')

    def __init__(self, java_object=None
                 , java_class_name: str = None
                 , python_shadow: Optional[FactoryManager] = None
                 , auto_load_java: bool = False
                 , name=None, simulator_type: Union[str, Type[FactorySimulator]] = FastFactorySimulator):
        super().__init__(name=name, simulator_type=simulator_type)
        self.java_awi = None
        if python_shadow is None:
            python_shadow_object = self
        else:
            python_shadow_object = python_shadow
        self.python_shadow = python_shadow_object
        self.init_java_bridge(java_object=java_object, java_class_name=java_class_name, auto_load_java=auto_load_java
                              , python_shadow_object=python_shadow_object)
        if java_object is None:
            map = to_dict_for_java(self)
            map.pop(PYTHON_CLASS_IDENTIFIER, None)
            map['simulatorType'] = self.simulator_type.__class__.__name__
            self._java_object.construct(to_java(map))

    def getNegotiationRequests(self):
        return to_java(self.requested_negotiations)

    def getRunningNegotiations(self):
        return to_java(self.running_negotiations)

    def requestNegotiation(self, cfp: CFP, negotiator: Negotiator = None, ufun: UtilityFunction = None) -> bool:
        return self.request_negotiation(from_java(cfp), JavaSAONegotiator(negotiator, None)
                                        , JavaUtilityFunction(ufun, None))

    class Java:
        implements = ['jnegmas.apps.scml.factory_managers.FactoryManager']


class JavaDoNothingFactoryManager(JavaFactoryManager):
    def __init__(self, auto_load_java: bool = False
                 , name=None, simulator_type: Union[str, Type[FactorySimulator]] = FastFactorySimulator):
        super().__init__(name=name, simulator_type=simulator_type, auto_load_java=auto_load_java
                         , java_class_name='jnegmas.apps.scml.factory_managers.DoNothingFactoryManager')


class JavaGreedyFactoryManager(JavaFactoryManager):
    def __init__(self, auto_load_java: bool = False
                 , name=None, simulator_type: Union[str, Type[FactorySimulator]] = FastFactorySimulator):
        super().__init__(name=name, simulator_type=simulator_type, auto_load_java=auto_load_java
                         , java_class_name='jnegmas.apps.scml.factory_managers.GreedyFactoryManager'
                         , python_shadow=lambda: GreedyFactoryManager(name=name
                                                                      , simulator_type=self.simulator_type))


class JavaDummyMiddleMan(JavaFactoryManager):
    def __init__(self, auto_load_java: bool = False
                 , name=None, simulator_type: Union[str, Type[FactorySimulator]] = FastFactorySimulator):
        super().__init__(name=name, simulator_type=simulator_type, auto_load_java=auto_load_java
                         , java_class_name='jnegmas.apps.scml.factory_managers.DummyMiddleMan')



import pytest

from negmas import *


def test_pareto_frontier_does_not_depend_on_order():
    u1 = [0.5337723805661662, 0.8532272031479199, 0.4781281413197942, 0.7242899747791032, 0.3461879818432919
        , 0.2608677043479706, 0.9419131964655383, 0.29368079952747694, 0.6093201983562316, 0.7066918086398718]
    u2 = [0.0, 0.0, 1.0, 1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0]
    welfare = [_1 + _2 for _1, _2 in zip(u1, u2)]
    assert (welfare.index(max(welfare)) == 3)

    f1 = MappingUtilityFunction(lambda o: u1[o[0]])
    f2 = MappingUtilityFunction(lambda o: u2[o[0]])
    assert all(f1((i,)) == u1[i] for i in range(10))
    assert all(f2((i,)) == u2[i] for i in range(10))
    p1, l1 = pareto_frontier([f1, f2], outcomes=[(_,) for _ in range(10)])
    p2, l2 = pareto_frontier([f2, f1], outcomes=[(_,) for _ in range(10)])

    assert p1 == [(0.9419131964655383, 0.0), (0.7242899747791032, 1.0)]
    assert p2 == [(1.0, 0.7242899747791032), (0.0, 0.9419131964655383)]
    assert l1 == [6, 3]
    assert l2 == list(reversed(l1))
    assert len(p1) == len(p2)

    # reverse order of p2
    p2 = [(_[1], _[0]) for _ in p2]
    for a in p1:
        assert a in p2


def test_linear_utility():
    buyer_utility = LinearUtilityAggregationFunction({'cost': lambda x: - x
                                                         , 'number of items': lambda x: 0.5 * x
                                                         , 'delivery': {'delivered': 10.0, 'not delivered': -2.0}})
    assert buyer_utility({'cost': 1.0, 'number of items': 3, 'delivery': 'not delivered'}) == -1.0 + 1.5 - 2.0


def test_hypervolume_utility():
    f = HyperRectangleUtilityFunction(outcome_ranges=[None,
                                                      {0: (1.0, 2.0), 1: (1.0, 2.0)},
                                                      {0: (1.4, 2.0), 2: (2.0, 3.0)}]
                                      , utilities=[5.0, 2.0, lambda x: 2 * x[2] + x[0]])
    f_ignore_input = HyperRectangleUtilityFunction(outcome_ranges=[None,
                                                                   {0: (1.0, 2.0), 1: (1.0, 2.0)},
                                                                   {0: (1.4, 2.0), 2: (2.0, 3.0)}]
                                                   , utilities=[5.0, 2.0, lambda x: 2 * x[2] + x[0]]
                                                   , ignore_issues_not_in_input=True)
    f_ignore_failing = HyperRectangleUtilityFunction(outcome_ranges=[None,
                                                                     {0: (1.0, 2.0), 1: (1.0, 2.0)},
                                                                     {0: (1.4, 2.0), 2: (2.0, 3.0)}]
                                                     , utilities=[5.0, 2.0, lambda x: 2 * x[2] + x[0]]
                                                     , ignore_failing_range_utilities=True)
    f_ignore_both = HyperRectangleUtilityFunction(outcome_ranges=[None,
                                                                  {0: (1.0, 2.0), 1: (1.0, 2.0)},
                                                                  {0: (1.4, 2.0), 2: (2.0, 3.0)}]
                                                  , utilities=[5.0, 2.0, lambda x: 2 * x[2] + x[0]]
                                                  , ignore_failing_range_utilities=True
                                                  , ignore_issues_not_in_input=True)

    g = HyperRectangleUtilityFunction(outcome_ranges=[
        {0: (1.0, 2.0), 1: (1.0, 2.0)},
        {0: (1.4, 2.0), 2: (2.0, 3.0)}]
        , utilities=[2.0, lambda x: 2 * x[2] + x[0]])
    g_ignore_input = HyperRectangleUtilityFunction(outcome_ranges=[
        {0: (1.0, 2.0), 1: (1.0, 2.0)},
        {0: (1.4, 2.0), 2: (2.0, 3.0)}]
        , utilities=[2.0, lambda x: 2 * x[2] + x[0]]
        , ignore_issues_not_in_input=True)
    g_ignore_failing = HyperRectangleUtilityFunction(outcome_ranges=[
        {0: (1.0, 2.0), 1: (1.0, 2.0)},
        {0: (1.4, 2.0), 2: (2.0, 3.0)}]
        , utilities=[2.0, lambda x: 2 * x[2] + x[0]]
        , ignore_failing_range_utilities=True)
    g_ignore_both = HyperRectangleUtilityFunction(outcome_ranges=[
        {0: (1.0, 2.0), 1: (1.0, 2.0)},
        {0: (1.4, 2.0), 2: (2.0, 3.0)}]
        , utilities=[2.0, lambda x: 2 * x[2] + x[0]]
        , ignore_failing_range_utilities=True
        , ignore_issues_not_in_input=True)

    funs = [g, g_ignore_input, g_ignore_failing, g_ignore_both
        , f, f_ignore_input, f_ignore_failing, f_ignore_both]
    outcomes = [[1.5, 1.5, 2.5],  # belongs to all volumes
                [1.5, 1.5, 1.0],  # belongs to first
                {0: 1.5, 2: 2.5},
                {0: 11.5, 1: 11.5, 2: 12.5},
                [1.5],
                {2: 2.5},
                ]
    expected = [
        [8.5, 8.5, 8.5, 8.5, 13.5, 13.5, 13.5, 13.5],
        [2.0, 2.0, 2.0, 2.0, 7.0, 7.0, 7.0, 7.0],
        [None, 6.5, None, 6.5, None, 11.5, None, 11.5],
        [0.0, 0.0, 0.0, 0.0, 5.0, 5.0, 5.0, 5.0],
        [None, 0.0, None, 0.0, None, 5.0, None, 5.0],
        [None, 0.0, None, 0.0, None, 5.0, None, 5.0],
    ]

    for outcome, expectation in zip(outcomes, expected):
        utilities = [f(outcome) for f in funs]
        for i, (u, e) in enumerate(zip(utilities, expectation)):
            # print(i, utilities, outcome)
            assert u == e


if __name__ == '__main__':
    pytest.main(args=[__file__])
